from functools import wraps
from types import TracebackType
from typing import (
    ClassVar,
    Optional,
    Sequence,
    List,
    Dict,
    Type,
    Iterable,
    Any,
    IO,
    Tuple,
    cast,
)

from dlt.common.configuration.exceptions import ConfigurationValueError
from dlt.common.destination.exceptions import (
    DestinationUndefinedEntity,
    DestinationTransientException,
    DestinationTerminalException,
)

import weaviate
from weaviate.classes.config import (
    Configure,
    Property,
    DataType,
    Tokenization,
)
from weaviate.classes.query import Filter, Sort
from weaviate.util import generate_uuid5
from weaviate.exceptions import (
    WeaviateConnectionError,
    WeaviateClosedClientError,
    WeaviateInvalidInputError,
    UnexpectedStatusCodeError,
)

from dlt.common import logger
from dlt.common.json import json
from dlt.common.pendulum import pendulum
from dlt.common.typing import StrAny, TFun
from dlt.common.time import ensure_pendulum_datetime_utc
from dlt.common.schema import Schema, TSchemaTables, TTableSchemaColumns
from dlt.common.schema.typing import (
    C_DLT_LOAD_ID,
    C_DLT_LOADS_TABLE_LOAD_ID,
    TColumnSchema,
    TColumnType,
)
from dlt.common.schema.utils import (
    get_columns_names_with_prop,
    loads_table,
    normalize_table_identifiers,
    version_table,
)
from dlt.common.destination import DestinationCapabilitiesContext
from dlt.common.destination.client import (
    PreparedTableSchema,
    RunnableLoadJob,
    JobClientBase,
    WithStateSync,
    LoadJob,
)
from dlt.common.storages import FileStorage

from dlt.destinations.impl.weaviate.weaviate_adapter import VECTORIZE_HINT, TOKENIZATION_HINT
from dlt.destinations.job_client_impl import StorageSchemaInfo, StateInfo
from dlt.destinations.impl.weaviate.configuration import WeaviateClientConfiguration
from dlt.destinations.impl.weaviate.exceptions import PropertyNameConflict, WeaviateBatchError
from dlt.destinations.utils import get_pipeline_state_query_columns


# Type mapping from Weaviate type strings (output of WeaviateTypeMapper) to v4 DataType enums
# The WeaviateTypeMapper in factory.py converts dlt types to these Weaviate type strings:
#   dlt type -> Weaviate string -> DataType enum
#   text -> "text" -> TEXT
#   double -> "number" -> NUMBER
#   bool -> "boolean" -> BOOL
#   timestamp/date -> "date" -> DATE
#   bigint -> "int" -> INT
#   binary -> "blob" -> BLOB (binary data like images, PDFs, etc.)
#   json -> "text" -> TEXT (complex nested objects are JSON-serialized)
#   decimal -> "text" -> TEXT (decimal precision preserved as string)
#   wei -> "number" -> NUMBER (Ethereum currency values as floats)
#   time -> "text" -> TEXT (time values as ISO strings)
WEAVIATE_TYPE_TO_DATATYPE: Dict[str, DataType] = {
    "text": DataType.TEXT,
    "number": DataType.NUMBER,
    "boolean": DataType.BOOL,
    "date": DataType.DATE,
    "int": DataType.INT,
    "blob": DataType.BLOB,
}


def wrap_weaviate_error(f: TFun) -> TFun:
    @wraps(f)
    def _wrap(self: JobClientBase, *args: Any, **kwargs: Any) -> Any:
        try:
            return f(self, *args, **kwargs)
        except WeaviateInvalidInputError as term_ex:
            raise DestinationTerminalException(term_ex) from term_ex
        except UnexpectedStatusCodeError as status_ex:
            # special handling for non existing objects/collections
            if status_ex.status_code == 404:
                raise DestinationUndefinedEntity(status_ex)
            if status_ex.status_code == 403:
                raise DestinationTerminalException(status_ex)
            if status_ex.status_code == 422:
                if "conflict for property" in str(status_ex) or "none vectorizer module" in str(
                    status_ex
                ):
                    raise PropertyNameConflict(str(status_ex))
                raise DestinationTerminalException(status_ex)
            # looks like there are no more terminal exceptions
            raise DestinationTransientException(status_ex)
        except (WeaviateConnectionError, WeaviateClosedClientError) as conn_ex:
            raise DestinationTransientException(conn_ex)

    return _wrap  # type: ignore


def wrap_batch_error(f: TFun) -> TFun:
    @wraps(f)
    def _wrap(*args: Any, **kwargs: Any) -> Any:
        try:
            return f(*args, **kwargs)
        except WeaviateBatchError as batch_ex:
            errors = batch_ex.args[0]
            message = str(errors)
            # TODO: actually put the job in failed/retry state and prepare exception message with full info on failing item
            if "invalid" in message and "property" in message:
                raise DestinationTerminalException(
                    f"Batch failed `{errors}` AND WILL **NOT** BE RETRIED"
                )
            if "conflict for property" in message:
                raise PropertyNameConflict(message)
            # v4 API gives different error for case-sensitive property conflicts
            if "no such prop with name" in message.lower() and "in class" in message.lower():
                raise PropertyNameConflict(message)
            raise DestinationTransientException(f"Batch failed `{errors}` AND WILL BE RETRIED")
        except Exception:
            raise DestinationTransientException("Batch failed AND WILL BE RETRIED")

    return _wrap  # type: ignore


class LoadWeaviateJob(RunnableLoadJob):
    def __init__(
        self,
        file_path: str,
        collection_name: str,
    ) -> None:
        super().__init__(file_path)
        self._job_client: WeaviateClient = None
        self._collection_name = collection_name

    def run(self) -> None:
        self._db_client = self._job_client.db_client
        self._client_config = self._job_client.config
        self.unique_identifiers = self.list_unique_identifiers(self._load_table)
        self.nested_indices = [
            i
            for i, field in self._schema.get_table_columns(self.load_table_name).items()
            if field["data_type"] == "json"
        ]
        self.date_indices = [
            i
            for i, field in self._schema.get_table_columns(self.load_table_name).items()
            if field["data_type"] == "date"
        ]
        with FileStorage.open_zipsafe_ro(self._file_path) as f:
            self.load_batch(f)

    @wrap_weaviate_error
    def load_batch(self, f: IO[str]) -> None:
        """Load all the lines from stream `f` using Weaviate batch.
        Weaviate batch supports retries so we do not need to do that.
        """
        collection = self._db_client.collections.get(self._collection_name)

        @wrap_batch_error
        def check_batch_result(failed_objects: List[Any]) -> None:
            """Check for failed objects and raise error if any"""
            if failed_objects:
                # Collect error messages from failed objects
                errors = []
                for obj in failed_objects:
                    if hasattr(obj, "message"):
                        errors.append(obj.message)
                    else:
                        errors.append(str(obj))
                if errors:
                    raise WeaviateBatchError({"error": [{"message": err} for err in errors]})

        # Collect all objects to insert
        objects_to_insert = []
        for line in f:
            data = json.loads(line)
            # serialize json types
            for key in self.nested_indices:
                if key in data:
                    data[key] = json.dumps(data[key])
            for key in self.date_indices:
                if key in data:
                    data[key] = ensure_pendulum_datetime_utc(data[key]).isoformat()
            if self.unique_identifiers:
                uuid = self.generate_uuid(data, self.unique_identifiers, self._collection_name)
            else:
                uuid = None

            objects_to_insert.append({"properties": data, "uuid": uuid})

        # Use batch insert with the v4 API
        with collection.batch.dynamic() as batch:
            for obj in objects_to_insert:
                batch.add_object(
                    properties=obj["properties"],
                    uuid=obj["uuid"],
                )

        # Check for failed objects
        failed_objects = collection.batch.failed_objects
        if failed_objects:
            check_batch_result(failed_objects)

    def list_unique_identifiers(self, table_schema: PreparedTableSchema) -> Sequence[str]:
        if table_schema.get("write_disposition") == "merge":
            primary_keys = get_columns_names_with_prop(table_schema, "primary_key")
            if primary_keys:
                return primary_keys
        return get_columns_names_with_prop(table_schema, "unique")

    def generate_uuid(
        self, data: Dict[str, Any], unique_identifiers: Sequence[str], collection_name: str
    ) -> str:
        data_id = "_".join([str(data[key]) for key in unique_identifiers])
        return str(generate_uuid5(data_id, collection_name))


class WeaviateClient(JobClientBase, WithStateSync):
    """Weaviate client implementation using v4 API."""

    def __init__(
        self,
        schema: Schema,
        config: WeaviateClientConfiguration,
        capabilities: DestinationCapabilitiesContext,
    ) -> None:
        super().__init__(schema, config, capabilities)
        # get definitions of the dlt tables, normalize column names and keep for later use
        version_table_ = normalize_table_identifiers(version_table(), schema.naming)
        self.version_collection_properties = list(version_table_["columns"].keys())
        loads_table_ = normalize_table_identifiers(loads_table(), schema.naming)
        self.loads_collection_properties = list(loads_table_["columns"].keys())
        state_table_ = normalize_table_identifiers(
            get_pipeline_state_query_columns(), schema.naming
        )
        self.pipeline_state_properties = list(state_table_["columns"].keys())

        self.config: WeaviateClientConfiguration = config
        self.db_client: weaviate.WeaviateClient = None

        self._vectorizer_config = config.vectorizer
        self._module_config = config.module_config
        self.type_mapper = self.capabilities.get_type_mapper()

    @property
    def dataset_name(self) -> str:
        return self.config.normalize_dataset_name(self.schema)

    @property
    def sentinel_collection(self) -> str:
        # if no dataset name is provided we still want to create sentinel collection
        return self.dataset_name or "DltSentinelClass"

    @staticmethod
    def create_db_client(config: WeaviateClientConfiguration) -> weaviate.WeaviateClient:
        """Create a Weaviate client using v4 API.

        Supports three connection types:
        - "cloud": For Weaviate Cloud Services (uses connect_to_weaviate_cloud)
        - "local": For local Docker instances (uses connect_to_local)
        - "custom": For self-hosted instances (uses connect_to_custom)

        If connection_type is not specified, it's auto-detected from the URL:
        - URLs containing ".weaviate.cloud" -> "cloud"
        - URLs with "localhost" or "127.0.0.1" -> "local"
        - Other URLs -> "custom" (requires http_port and grpc_port)
        """
        url = config.credentials.url
        api_key = config.credentials.api_key
        headers = config.credentials.additional_headers or {}
        connection_type = config.connection_type

        # Create auth config if API key is provided
        auth_config = None
        if api_key:
            auth_config = weaviate.auth.AuthApiKey(api_key)

        # Auto-detect connection type if not specified
        if connection_type is None:
            if ".weaviate.cloud" in url or ".wcs.api.weaviate.io" in url:
                connection_type = "cloud"
            elif "localhost" in url or "127.0.0.1" in url:
                connection_type = "local"
            else:
                connection_type = "custom"

        if connection_type == "cloud":
            # Use connect_to_weaviate_cloud for Weaviate Cloud Services
            # Ensure URL has https:// prefix
            cluster_url = url if url.startswith("https://") else f"https://{url}"
            return weaviate.connect_to_weaviate_cloud(
                cluster_url=cluster_url,
                auth_credentials=auth_config,
                headers=headers,
                skip_init_checks=True,
            )
        elif connection_type == "local":
            # Use connect_to_local for local Docker instances
            # Parse host from URL
            host = url.replace("http://", "").replace("https://", "").split(":")[0]
            http_port = config.credentials.http_port or 8080
            grpc_port = config.credentials.grpc_port or 50051
            return weaviate.connect_to_local(
                host=host,
                port=http_port,
                grpc_port=grpc_port,
                headers=headers,
                auth_credentials=auth_config,
            )
        else:  # custom
            # Use connect_to_custom for self-hosted instances
            # Require explicit ports for custom connections
            http_port = config.credentials.http_port
            grpc_port = config.credentials.grpc_port
            if http_port is None or grpc_port is None:
                raise ConfigurationValueError(
                    "http_port and grpc_port",
                    "http_port and grpc_port are required when connection_type is 'custom'. "
                    "Set them in [destination.weaviate.credentials] or use connection_type='local' "
                    "for default ports (http: 8080, grpc: 50051).",
                )
            host = url.replace("http://", "").replace("https://", "").split(":")[0]
            is_secure = url.startswith("https")
            return weaviate.connect_to_custom(
                http_host=host,
                http_port=http_port,
                http_secure=is_secure,
                grpc_host=host,
                grpc_port=grpc_port,
                grpc_secure=is_secure,
                auth_credentials=auth_config,
                headers=headers,
                skip_init_checks=True,
            )

    def make_qualified_collection_name(self, table_name: str) -> str:
        """Make a full Weaviate collection name from a table name by prepending
        the dataset name if it exists.
        """
        dataset_separator = self.config.dataset_separator

        return (
            f"{self.dataset_name}{dataset_separator}{table_name}"
            if self.dataset_name
            else table_name
        )

    def get_collection_schema(self, table_name: str) -> Dict[str, Any]:
        """Get the Weaviate collection schema for a table."""
        collection_name = self.make_qualified_collection_name(table_name)
        try:
            collection = self.db_client.collections.get(collection_name)
            config = collection.config.get()
            return self._config_to_dict(config)
        except UnexpectedStatusCodeError as e:
            if e.status_code == 404:
                raise
            raise
        except Exception as e:
            if "404" in str(e) or "not found" in str(e).lower():
                # Re-raise as UnexpectedStatusCodeError with status_code attribute
                err = UnexpectedStatusCodeError(f"Collection {collection_name} not found", None)
                raise err from e
            raise

    def _config_to_dict(self, config: Any) -> Dict[str, Any]:
        """Convert a collection config to a dictionary format similar to v3"""
        # Get vectorizer name - in v4 it's directly a Vectorizers enum
        vectorizer_name = "none"
        if config.vectorizer:
            if hasattr(config.vectorizer, "value"):
                vectorizer_name = config.vectorizer.value
            else:
                vectorizer_name = str(config.vectorizer)

        result: Dict[str, Any] = {
            "class": config.name,
            "properties": [],
            "vectorizer": vectorizer_name,
        }

        # Add vectorIndexConfig if available
        if config.vector_index_config:
            result["vectorIndexConfig"] = {
                "skip": getattr(config.vector_index_config, "skip", False),
            }

        # Add properties
        for prop in config.properties:
            # Get data type - handle both enum and string
            data_type_value = (
                prop.data_type.value if hasattr(prop.data_type, "value") else str(prop.data_type)
            )
            # Clean up data type value if it looks like "DataType.TEXT" -> "text"
            if "." in data_type_value:
                data_type_value = data_type_value.split(".")[-1].lower()

            prop_dict: Dict[str, Any] = {
                "name": prop.name,
                "dataType": [data_type_value],
            }
            if prop.tokenization:
                prop_dict["tokenization"] = (
                    prop.tokenization.value
                    if hasattr(prop.tokenization, "value")
                    else str(prop.tokenization)
                )

            # Add moduleConfig for vectorization skip info
            # In v4 API, the skip value is in prop.vectorizer_config.skip
            if vectorizer_name != "none":
                skip_vectorization = False
                if hasattr(prop, "vectorizer_config") and prop.vectorizer_config:
                    skip_vectorization = getattr(prop.vectorizer_config, "skip", False)
                prop_dict["moduleConfig"] = {
                    vectorizer_name: {
                        "skip": skip_vectorization,
                    }
                }
            result["properties"].append(prop_dict)

        return result

    def create_collection(
        self, collection_config: Dict[str, Any], full_collection_name: Optional[str] = None
    ) -> None:
        """Create a Weaviate collection.

        Args:
            collection_config: The collection configuration to create.
            full_collection_name: The full name of the collection to create. If not
                provided, the collection name will be prepended with the dataset name
                if it exists.
        """
        name = (
            self.make_qualified_collection_name(collection_config["class"])
            if full_collection_name is None
            else full_collection_name
        )

        # Build properties
        properties = []
        for prop in collection_config.get("properties", []):
            prop_config = self._make_property_config(prop)
            properties.append(prop_config)

        # Determine vectorizer configuration
        vectorizer = collection_config.get("vectorizer", "none")
        vectorizer_config = None

        if vectorizer == "none":
            # No vectorization
            vectorizer_config = Configure.Vectorizer.none()
        else:
            # Use the specified vectorizer
            vectorizer_config = self._get_vectorizer_config(vectorizer)

        # Create the collection
        self.db_client.collections.create(
            name=name,
            properties=properties,
            vectorizer_config=vectorizer_config,
        )

    def _make_property_config(self, prop: Dict[str, Any]) -> Property:
        """Create a Property config from a property dict"""
        data_type_str = (
            prop["dataType"][0] if isinstance(prop["dataType"], list) else prop["dataType"]
        )
        data_type = WEAVIATE_TYPE_TO_DATATYPE.get(data_type_str.lower(), DataType.TEXT)

        # Handle tokenization
        tokenization = None
        if "tokenization" in prop:
            token_map = {
                "word": Tokenization.WORD,
                "lowercase": Tokenization.LOWERCASE,
                "whitespace": Tokenization.WHITESPACE,
                "field": Tokenization.FIELD,
            }
            tokenization = token_map.get(prop["tokenization"])

        # Handle skip vectorization - this is per-property in v4
        skip_vectorization = False
        if "moduleConfig" in prop:
            for _module_name, module_config in prop["moduleConfig"].items():
                if module_config.get("skip", False):
                    skip_vectorization = True
                    break

        return Property(
            name=prop["name"],
            data_type=data_type,
            tokenization=tokenization,
            skip_vectorization=skip_vectorization,
            # Disable property name vectorization to avoid contextionary issues with
            # property names that contain underscores or are not English words
            vectorize_property_name=False,
        )

    def _get_vectorizer_config(self, vectorizer: str) -> Any:
        """Get the vectorizer configuration based on vectorizer name"""
        # Map vectorizer names to Configure.Vectorizer methods
        vectorizer_map = {
            "text2vec-openai": Configure.Vectorizer.text2vec_openai,
            "text2vec-cohere": Configure.Vectorizer.text2vec_cohere,
            "text2vec-contextionary": Configure.Vectorizer.text2vec_contextionary,
            "text2vec-huggingface": Configure.Vectorizer.text2vec_huggingface,
            "text2vec-palm": Configure.Vectorizer.text2vec_palm,
        }

        if vectorizer in vectorizer_map:
            # Get the module config if available
            module_conf = self._module_config.get(vectorizer, {}) if self._module_config else {}
            try:
                # Filter out incompatible config keys for each vectorizer
                if vectorizer == "text2vec-contextionary":
                    # Contextionary accepts vectorize_collection_name
                    filtered_conf = {
                        k: v
                        for k, v in module_conf.items()
                        if k in ["vectorize_collection_name", "vectorizeClassName"]
                    }
                    # Map old v3 config key names to v4
                    if "vectorizeClassName" in filtered_conf:
                        filtered_conf["vectorize_collection_name"] = filtered_conf.pop(
                            "vectorizeClassName"
                        )
                    return (
                        vectorizer_map[vectorizer](**filtered_conf)
                        if filtered_conf
                        else vectorizer_map[vectorizer]()
                    )
                else:
                    return vectorizer_map[vectorizer](**module_conf)
            except TypeError:
                # If module_conf keys don't match, try without config
                return vectorizer_map[vectorizer]()

        # Default to none if not found
        return Configure.Vectorizer.none()

    def add_property_to_collection(self, collection_name: str, prop_schema: Dict[str, Any]) -> None:
        """Add a property to an existing Weaviate collection.

        Args:
            collection_name: The name of the collection to add the property to.
            prop_schema: The property schema to add.
        """
        full_name = self.make_qualified_collection_name(collection_name)
        collection = self.db_client.collections.get(full_name)
        prop_config = self._make_property_config(prop_schema)
        collection.config.add_property(prop_config)

    def delete_collection(self, collection_name: str) -> None:
        """Delete a Weaviate collection.

        Args:
            collection_name: The name of the collection to delete.
        """
        self.db_client.collections.delete(self.make_qualified_collection_name(collection_name))

    def delete_all_collections(self) -> None:
        """Delete all Weaviate collections from Weaviate instance and all data
        associated with it.
        """
        self.db_client.collections.delete_all()

    def query_collection(
        self, collection_name: str, properties: List[str], limit: int = 100
    ) -> List[Dict[str, Any]]:
        """Query a Weaviate collection.

        Args:
            collection_name: The name of the collection to query.
            properties: The properties to return.
            limit: Maximum number of results to return.

        Returns:
            A list of objects from the collection.
        """
        full_name = self.make_qualified_collection_name(collection_name)
        collection = self.db_client.collections.get(full_name)
        response = collection.query.fetch_objects(
            limit=limit,
            return_properties=properties,
        )
        return [{"properties": obj.properties, "uuid": str(obj.uuid)} for obj in response.objects]

    def create_object(self, obj: Dict[str, Any], collection_name: str) -> None:
        """Create a Weaviate object.

        Args:
            obj: The object to create.
            collection_name: The name of the collection to create the object in.
        """
        full_name = self.make_qualified_collection_name(collection_name)
        collection = self.db_client.collections.get(full_name)
        collection.data.insert(properties=obj)

    def drop_storage(self) -> None:
        """Drop the dataset from Weaviate instance.

        Deletes all collections in the dataset and all data associated with them.
        Deletes the sentinel collection as well.

        If dataset name was not provided, it deletes all the tables in the current schema
        """
        collections = self.db_client.collections.list_all()
        collection_names = list(collections.keys())

        if self.dataset_name:
            prefix = f"{self.dataset_name}{self.config.dataset_separator}"

            for collection_name in collection_names:
                if collection_name.startswith(prefix):
                    self.db_client.collections.delete(collection_name)
        else:
            # in case of no dataset prefix do our best and delete all tables in the schema
            for table_name in self.schema.tables.keys():
                if table_name in collection_names:
                    self.db_client.collections.delete(table_name)

        self._delete_sentinel_collection()

    @wrap_weaviate_error
    def initialize_storage(self, truncate_tables: Iterable[str] = None) -> None:
        if not self.is_storage_initialized():
            self._create_sentinel_collection()
        elif truncate_tables:
            for table_name in truncate_tables:
                try:
                    collection_schema = self.get_collection_schema(table_name)
                except UnexpectedStatusCodeError as e:
                    if hasattr(e, "status_code") and e.status_code == 404:
                        continue
                    raise
                except Exception as e:
                    if "404" in str(e) or "not found" in str(e).lower():
                        continue
                    raise

                self.delete_collection(table_name)
                self.create_collection(
                    collection_schema, full_collection_name=collection_schema["class"]
                )

    @wrap_weaviate_error
    def is_storage_initialized(self) -> bool:
        try:
            collections = self.db_client.collections.list_all()
            return self.sentinel_collection in collections
        except Exception:
            return False

    def _create_sentinel_collection(self) -> None:
        """Create an empty collection to indicate that the storage is initialized."""
        self.db_client.collections.create(
            name=self.sentinel_collection,
            vectorizer_config=Configure.Vectorizer.none(),
        )

    def _delete_sentinel_collection(self) -> None:
        """Delete the sentinel collection."""
        try:
            self.db_client.collections.delete(self.sentinel_collection)
        except Exception:
            pass  # Ignore if not found

    @wrap_weaviate_error
    def update_stored_schema(
        self,
        only_tables: Iterable[str] = None,
        expected_update: TSchemaTables = None,
    ) -> Optional[TSchemaTables]:
        applied_update = super().update_stored_schema(only_tables, expected_update)
        # Retrieve the schema from Weaviate
        try:
            schema_info = self.get_stored_schema_by_hash(self.schema.stored_version_hash)
        except DestinationUndefinedEntity:
            schema_info = None
        if schema_info is None:
            logger.info(
                f"Schema with hash {self.schema.stored_version_hash} "
                "not found in the storage. upgrading"
            )
            # TODO: return a real updated table schema (like in SQL job client)
            self._execute_schema_update(only_tables)
        else:
            logger.info(
                f"Schema with hash {self.schema.stored_version_hash} "
                f"inserted at {schema_info.inserted_at} found "
                "in storage, no upgrade required"
            )

        return applied_update

    def _execute_schema_update(self, only_tables: Iterable[str]) -> None:
        for table_name in only_tables or self.schema.tables.keys():
            exists, existing_columns = self.get_storage_table(table_name)
            # TODO: detect columns where vectorization was added or removed and modify it. currently we ignore change of hints
            new_columns = self.schema.get_new_table_columns(
                table_name,
                existing_columns,
                case_sensitive=self.capabilities.generates_case_sensitive_identifiers(),
            )
            logger.info(f"Found {len(new_columns)} updates for {table_name} in {self.schema.name}")
            if len(new_columns) > 0:
                if exists:
                    is_collection_vectorized = self._is_collection_vectorized(table_name)
                    for column in new_columns:
                        prop = self._make_property_schema(
                            column["name"], column, is_collection_vectorized
                        )
                        self.add_property_to_collection(table_name, prop)
                else:
                    collection_config = self.make_weaviate_collection_schema(table_name)
                    self.create_collection(collection_config)
        self._update_schema_in_storage(self.schema)

    def get_storage_table(self, table_name: str) -> Tuple[bool, TTableSchemaColumns]:
        table_schema: TTableSchemaColumns = {}

        try:
            collection_schema = self.get_collection_schema(table_name)
        except UnexpectedStatusCodeError as e:
            if hasattr(e, "status_code") and e.status_code == 404:
                return False, table_schema
            raise
        except Exception as e:
            if "404" in str(e) or "not found" in str(e).lower():
                return False, table_schema
            raise

        # Convert Weaviate collection schema to dlt table schema
        for prop in collection_schema["properties"]:
            schema_c: TColumnSchema = {
                "name": prop["name"],
                **self._from_db_type(prop["dataType"][0], None, None),
            }
            table_schema[prop["name"]] = schema_c
        return True, table_schema

    def get_stored_state(self, pipeline_name: str) -> Optional[StateInfo]:
        """Loads compressed state from destination storage"""
        # normalize properties
        p_load_id = self.schema.naming.normalize_identifier(C_DLT_LOADS_TABLE_LOAD_ID)
        p_dlt_load_id = self.schema.naming.normalize_identifier(C_DLT_LOAD_ID)
        p_pipeline_name = self.schema.naming.normalize_identifier("pipeline_name")
        p_status = self.schema.naming.normalize_identifier("status")

        # we need to find a stored state that matches a load id that was completed
        # we retrieve the state in blocks of 10 for this
        stepsize = 10
        offset = 0
        while True:
            state_records = self.get_records(
                self.schema.state_table_name,
                # search by package load id which is guaranteed to increase over time
                sort_by=p_dlt_load_id,
                sort_order="desc",
                filter_field=p_pipeline_name,
                filter_value=pipeline_name,
                limit=stepsize,
                offset=offset,
                properties=self.pipeline_state_properties,
            )
            offset += stepsize
            if len(state_records) == 0:
                return None
            for state in state_records:
                load_id = state[p_dlt_load_id]
                load_records = self.get_records(
                    self.schema.loads_table_name,
                    filter_field=p_load_id,
                    filter_value=load_id,
                    limit=1,
                    properties=[p_load_id, p_status],
                )
                # if there is a load for this state which was successful, return the state
                if len(load_records):
                    return StateInfo(**state)

    def get_stored_schema(self, schema_name: str = None) -> Optional[StorageSchemaInfo]:
        """Retrieves newest schema from destination storage"""
        p_schema_name = self.schema.naming.normalize_identifier("schema_name")
        p_inserted_at = self.schema.naming.normalize_identifier("inserted_at")

        try:
            records = self.get_records(
                self.schema.version_table_name,
                sort_by=p_inserted_at,
                sort_order="desc",
                filter_field=p_schema_name if schema_name else None,
                filter_value=schema_name if schema_name else None,
                limit=1,
            )
            if records:
                return StorageSchemaInfo(**records[0])
            return None
        except Exception:
            return None

    def get_stored_schema_by_hash(self, schema_hash: str) -> Optional[StorageSchemaInfo]:
        p_version_hash = self.schema.naming.normalize_identifier("version_hash")
        try:
            records = self.get_records(
                self.schema.version_table_name,
                filter_field=p_version_hash,
                filter_value=schema_hash,
                limit=1,
            )
            if records:
                return StorageSchemaInfo(**records[0])
            return None
        except Exception:
            return None

    @wrap_weaviate_error
    def get_records(
        self,
        table_name: str,
        filter_field: str = None,
        filter_value: Any = None,
        sort_by: str = None,
        sort_order: str = "asc",
        limit: int = 0,
        offset: int = 0,
        properties: List[str] = None,
    ) -> List[Dict[str, Any]]:
        """Get records from a collection using v4 API"""
        # fail if collection does not exist
        self.get_collection_schema(table_name)

        full_name = self.make_qualified_collection_name(table_name)
        collection = self.db_client.collections.get(full_name)

        # build query
        if not properties:
            properties = list(self.schema.get_table_columns(table_name).keys())

        # Build filter if provided
        filters = None
        if filter_field and filter_value is not None:
            filters = Filter.by_property(filter_field).equal(filter_value)

        # Build sort if provided
        sort = None
        if sort_by:
            if sort_order == "desc":
                sort = Sort.by_property(sort_by, ascending=False)
            else:
                sort = Sort.by_property(sort_by, ascending=True)

        # Execute query
        response = collection.query.fetch_objects(
            filters=filters,
            sort=sort,
            limit=limit if limit > 0 else None,
            offset=offset if offset > 0 else None,
            return_properties=properties,
        )

        records = [obj.properties for obj in response.objects]
        if records is None:
            raise DestinationTransientException(f"Could not obtain records for `{full_name}`")
        return cast(List[Dict[str, Any]], records)

    def make_weaviate_collection_schema(self, table_name: str) -> Dict[str, Any]:
        """Creates a Weaviate collection schema from a table schema."""
        collection_schema: Dict[str, Any] = {
            "class": table_name,
            "properties": self._make_properties(table_name),
        }

        # check if any column requires vectorization
        if self._is_collection_vectorized(table_name):
            collection_schema["vectorizer"] = self._vectorizer_config
            # Only include module config for the active vectorizer
            if self._module_config and self._vectorizer_config in self._module_config:
                collection_schema["moduleConfig"] = {
                    self._vectorizer_config: self._module_config[self._vectorizer_config]
                }
        else:
            collection_schema["vectorizer"] = "none"
            collection_schema["vectorIndexConfig"] = {"skip": True}

        return collection_schema

    def _is_collection_vectorized(self, table_name: str) -> bool:
        """Tells is any of the columns has vectorize hint set"""
        return (
            len(get_columns_names_with_prop(self.schema.get_table(table_name), VECTORIZE_HINT)) > 0
        )

    def _make_properties(self, table_name: str) -> List[Dict[str, Any]]:
        """Creates a Weaviate properties schema from a table schema.

        Args:
            table: The table name for which columns should be converted to properties
        """
        is_collection_vectorized = self._is_collection_vectorized(table_name)
        return [
            self._make_property_schema(column_name, column, is_collection_vectorized)
            for column_name, column in self.schema.get_table_columns(table_name).items()
        ]

    def _make_property_schema(
        self, column_name: str, column: TColumnSchema, is_collection_vectorized: bool
    ) -> Dict[str, Any]:
        extra_kv = {}

        vectorizer_name = self._vectorizer_config
        # x-weaviate-vectorize: (bool) means that this field should be vectorized
        if is_collection_vectorized and not column.get(VECTORIZE_HINT, False):
            # tell weaviate explicitly to not vectorize when column has no vectorize hint
            extra_kv["moduleConfig"] = {
                vectorizer_name: {
                    "skip": True,
                }
            }

        # x-weaviate-tokenization: (str) specifies the method to use
        # for tokenization
        if TOKENIZATION_HINT in column:
            extra_kv["tokenization"] = column[TOKENIZATION_HINT]  # type: ignore

        return {
            "name": column_name,
            "dataType": [self.type_mapper.to_destination_type(column, None)],
            **extra_kv,
        }

    def create_load_job(
        self, table: PreparedTableSchema, file_path: str, load_id: str, restore: bool = False
    ) -> LoadJob:
        return LoadWeaviateJob(
            file_path,
            collection_name=self.make_qualified_collection_name(table["name"]),
        )

    @wrap_weaviate_error
    def complete_load(self, load_id: str) -> None:
        # corresponds to order of the columns in loads_table()
        values = [
            load_id,
            self.schema.name,
            0,
            pendulum.now().isoformat(),
            self.schema.version_hash,
        ]
        assert len(values) == len(self.loads_collection_properties)
        properties = {k: v for k, v in zip(self.loads_collection_properties, values)}
        self.create_object(properties, self.schema.loads_table_name)

    def __enter__(self) -> "WeaviateClient":
        self.db_client = self.create_db_client(self.config)
        return self

    def __exit__(
        self,
        exc_type: Type[BaseException],
        exc_val: BaseException,
        exc_tb: TracebackType,
    ) -> None:
        if self.db_client:
            self.db_client.close()
            self.db_client = None

    def _update_schema_in_storage(self, schema: Schema) -> None:
        schema_str = json.dumps(schema.to_dict())
        # corresponds to order of the columns in version_table()
        values = [
            schema.version,
            schema.ENGINE_VERSION,
            str(pendulum.now().isoformat()),
            schema.name,
            schema.stored_version_hash,
            schema_str,
        ]
        assert len(values) == len(self.version_collection_properties)
        properties = {k: v for k, v in zip(self.version_collection_properties, values)}
        self.create_object(properties, self.schema.version_table_name)

    def _from_db_type(
        self, wt_t: str, precision: Optional[int], scale: Optional[int]
    ) -> TColumnType:
        return self.type_mapper.from_destination_type(wt_t, precision, scale)

    # Backwards compatibility aliases
    @property
    def sentinel_class(self) -> str:
        """Alias for sentinel_collection for backwards compatibility"""
        return self.sentinel_collection

    def make_qualified_class_name(self, table_name: str) -> str:
        """Alias for make_qualified_collection_name for backwards compatibility"""
        return self.make_qualified_collection_name(table_name)

    def get_class_schema(self, table_name: str) -> Dict[str, Any]:
        """Alias for get_collection_schema for backwards compatibility"""
        return self.get_collection_schema(table_name)

    def create_class(
        self, class_schema: Dict[str, Any], full_class_name: Optional[str] = None
    ) -> None:
        """Alias for create_collection for backwards compatibility"""
        return self.create_collection(class_schema, full_class_name)

    def create_class_property(self, class_name: str, prop_schema: Dict[str, Any]) -> None:
        """Alias for add_property_to_collection for backwards compatibility"""
        return self.add_property_to_collection(class_name, prop_schema)

    def delete_class(self, class_name: str) -> None:
        """Alias for delete_collection for backwards compatibility"""
        return self.delete_collection(class_name)

    def query_class(self, class_name: str, properties: List[str]) -> Any:
        """Query a collection and return results in v3-compatible format for backwards compatibility"""
        full_name = self.make_qualified_collection_name(class_name)
        collection = self.db_client.collections.get(full_name)

        # Return a wrapper that provides .do() method for compatibility
        class QueryWrapper:
            def __init__(self, coll: Any, props: List[str]) -> None:
                self._collection = coll
                self._properties = props
                self._filters: Optional[Any] = None
                self._sort: Optional[Any] = None
                self._limit: Optional[int] = None
                self._offset: Optional[int] = None

            def with_where(self, where: Dict[str, Any]) -> "QueryWrapper":
                # Convert v3 where format to v4 filter
                path = where.get("path", [])
                operator = where.get("operator", "Equal")
                value = (
                    where.get("valueString")
                    or where.get("valueNumber")
                    or where.get("valueBoolean")
                )
                if path and value is not None:
                    prop_name = path[0]
                    if operator == "Equal":
                        self._filters = Filter.by_property(prop_name).equal(value)
                return self

            def with_sort(self, sort: Dict[str, Any]) -> "QueryWrapper":
                path = sort.get("path", [])
                order = sort.get("order", "asc")
                if path:
                    if order == "desc":
                        self._sort = Sort.by_property(path[0], ascending=False)
                    else:
                        self._sort = Sort.by_property(path[0], ascending=True)
                return self

            def with_limit(self, limit: int) -> "QueryWrapper":
                self._limit = limit
                return self

            def with_offset(self, offset: int) -> "QueryWrapper":
                self._offset = offset
                return self

            def do(self) -> Dict[str, Any]:
                response = self._collection.query.fetch_objects(
                    filters=self._filters,
                    sort=self._sort,
                    limit=self._limit,
                    offset=self._offset,
                    return_properties=self._properties,
                )
                # Convert to v3 format
                objects = [obj.properties for obj in response.objects]
                return {"data": {"Get": {self._collection.name: objects}}}

        return QueryWrapper(collection, properties)



import dataclasses
import os
from pathlib import Path
from typing import Final, Optional, Any, Dict, ClassVar, List

from dlt.common.destination.configuration import CsvFormatConfiguration
from dlt.common.libs.cryptography import decode_private_key
from dlt.common.typing import TSecretStrValue
from dlt.common.configuration.specs import ConnectionStringCredentials
from dlt.common.configuration.exceptions import ConfigurationValueError
from dlt.common.configuration import configspec
from dlt.common.destination.client import DestinationClientDwhWithStagingConfiguration
from dlt.common.utils import digest128
from dlt.destinations.impl.snowflake.utils import (
    read_snowflake_session_token,
    snowflake_session_token_available,
)


SNOWFLAKE_APPLICATION_ID = "dltHub_dlt"


@configspec(init=False)
class SnowflakeCredentialsWithoutDefaults(ConnectionStringCredentials):
    drivername: Final[str] = dataclasses.field(default="snowflake", init=False, repr=False, compare=False)  # type: ignore[misc]
    database: str = None
    host: str = None
    """Snowflake account identifier, e.g. `kgiotue-wn98412`"""
    username: Optional[str] = None
    warehouse: Optional[str] = None
    role: Optional[str] = None
    authenticator: Optional[str] = None
    token: Optional[str] = None
    private_key: Optional[TSecretStrValue] = None
    private_key_path: Optional[str] = None
    private_key_passphrase: Optional[TSecretStrValue] = None
    application: Optional[str] = SNOWFLAKE_APPLICATION_ID

    _snowflake_host: Optional[str] = None
    """Snowflake account URL, e.g. https://kgiotue-wn98412.snowflakecomputing.com"""

    __config_gen_annotations__: ClassVar[List[str]] = [
        "username",
        "password",
        "warehouse",
        "role",
    ]
    __query_params__: ClassVar[List[str]] = [
        "warehouse",
        "role",
        "authenticator",
        "token",
        "private_key",
        "private_key_path",
        "private_key_passphrase",
    ]

    def parse_native_representation(self, native_value: Any) -> None:
        super().parse_native_representation(native_value)
        for param in self.__query_params__:
            if param in self.query:
                setattr(self, param, self.query.get(param))

    def on_resolved(self) -> None:
        if self.private_key_path:
            try:
                self.private_key = Path(self.private_key_path).read_text("ascii")
            except Exception:
                raise ValueError(
                    "Make sure that `private_key` in dlt recognized format is at"
                    f" `{self.private_key_path}`. Note that binary formats are not supported"
                )

        if self._is_snowflake_auth and not self.private_key:
            if not self.username or not self.password:
                raise ConfigurationValueError(
                    "`SnowflakeCredentials` requires `username` and `password` to be specified when"
                    " using `snowflake` (username/password) authentication."
                )

    def get_query(self) -> Dict[str, Any]:
        query = dict(super().get_query() or {})
        for param in self.__query_params__:
            if self.get(param, None) is not None:
                query[param] = self[param]
        return query

    def to_connector_params(self) -> Dict[str, Any]:
        # gather all params in query
        query = self.get_query()
        if self.private_key:
            query["private_key"] = decode_private_key(self.private_key, self.private_key_passphrase)

        # we do not want passphrase to be passed
        query.pop("private_key_passphrase", None)

        conn_params: Dict[str, Any] = dict(
            query,
            user=self.username,
            password=self.password,
            account=self.host,
            database=self.database,
        )

        if self.application != "" and "application" not in conn_params:
            conn_params["application"] = self.application

        # snowflake-connector-python fails if `host` is None, so we only set it if it's not None
        if self._snowflake_host:
            conn_params["host"] = self._snowflake_host

        return conn_params

    @property
    def _is_snowflake_auth(self) -> bool:
        return not self.authenticator or self.authenticator == "snowflake"


@configspec(init=False)
class SnowflakeCredentials(SnowflakeCredentialsWithoutDefaults):
    _use_snowflake_session_token: bool = False

    def on_partial(self) -> None:
        if self.authenticator == "oauth" and not self.host:
            self._from_snowflake_session_token()

        if not self.is_partial():
            self.resolve()

    def on_resolved(self) -> None:
        if self.authenticator == "oauth" and not self.token:
            self._from_snowflake_session_token()

    def get_query(self) -> Dict[str, Any]:
        self._ensure_fresh_token()
        return super().get_query()

    def _from_snowflake_session_token(self) -> None:
        if not snowflake_session_token_available():
            raise ConfigurationValueError(
                "Snowflake-provided OAuth token not available. `dlt` expects this token to be"
                " available when `authenticator` is set to `oauth` and either `token` or `host` are"
                " not provided. "
            )
        self.host = os.environ["SNOWFLAKE_ACCOUNT"]
        self._snowflake_host = os.environ["SNOWFLAKE_HOST"]
        self._use_snowflake_session_token = True

    def _ensure_fresh_token(self) -> None:
        if self._use_snowflake_session_token:
            self.token = read_snowflake_session_token()


@configspec
class SnowflakeClientConfiguration(DestinationClientDwhWithStagingConfiguration):
    destination_type: Final[str] = dataclasses.field(default="snowflake", init=False, repr=False, compare=False)  # type: ignore[misc]
    credentials: SnowflakeCredentials = None

    stage_name: Optional[str] = None
    """Use an existing named stage instead of the default. Default uses the implicit table stage per table"""
    keep_staged_files: bool = True
    """Whether to keep or delete the staged files after COPY INTO succeeds"""

    csv_format: Optional[CsvFormatConfiguration] = None
    """Optional csv format configuration"""

    query_tag: Optional[str] = None
    """A tag with placeholders to tag sessions executing jobs"""

    create_indexes: bool = False
    """Whether UNIQUE or PRIMARY KEY constrains should be created"""

    use_vectorized_scanner: bool = False
    """Whether to use or not use the vectorized scanner in COPY INTO"""

    enable_atomic_swap: bool = False
    """When enabled and using the `staging-optimized` replace strategy, tables will be atomically swapped using `ALTER TABLE ... SWAP WITH ...` instead of being cloned. This preserves permissions and enables zero-downtime loading."""

    use_decfloat: bool = False
    """Whether to use DECFLOAT type for unbound decimals instead of DECIMAL"""

    def fingerprint(self) -> str:
        """Returns a fingerprint of host part of a connection string"""
        if self.credentials and self.credentials.host:
            return digest128(self.credentials.host)
        return ""



import abc
import csv
from packaging.version import Version
from typing import (
    IO,
    TYPE_CHECKING,
    Any,
    Dict,
    List,
    Literal,
    Optional,
    Sequence,
    Tuple,
    Type,
    NamedTuple,
    TypeVar,
    cast,
)

from dlt.common.json import json
from dlt.common.configuration import with_config
from dlt.common.data_writers.exceptions import (
    SpecLookupFailed,
    DataWriterNotFound,
    FileFormatForItemFormatNotFound,
    FileSpecNotFound,
    InvalidDataItem,
)
from dlt.common.destination.configuration import (
    CsvFormatConfiguration,
    CsvQuoting,
    ParquetFormatConfiguration,
)
from dlt.common.destination import (
    DestinationCapabilitiesContext,
    TLoaderFileFormat,
    LOADER_FILE_FORMATS,
)
from dlt.common.exceptions import ValueErrorWithKnownValues
from dlt.common.metrics import DataWriterMetrics
from dlt.common.schema.typing import TTableSchemaColumns
from dlt.common.typing import StrAny, TDataItem, TDataItems

if TYPE_CHECKING:
    from dlt.common.libs.pyarrow import pyarrow as pa


TDataItemFormat = Literal["arrow", "object", "file", "model"]
TWriter = TypeVar("TWriter", bound="DataWriter")


class FileWriterSpec(NamedTuple):
    file_format: TLoaderFileFormat
    """format of the output file"""
    data_item_format: TDataItemFormat
    """format of the input data"""
    file_extension: str
    is_binary_format: bool
    supports_schema_changes: Literal["True", "Buffer", "False"]
    """File format supports changes of schema: True - at any moment, Buffer - in memory buffer before opening file, False - not at all"""
    requires_destination_capabilities: bool = False
    supports_compression: bool = False
    file_max_items: Optional[int] = None
    """Set an upper limit on the number of items in one file"""


EMPTY_DATA_WRITER_METRICS = DataWriterMetrics("", 0, 0, 2**32, 0.0)


class DataWriter(abc.ABC):
    def __init__(self, f: IO[Any], caps: DestinationCapabilitiesContext = None) -> None:
        self._f = f
        self._caps = caps
        self.items_count = 0

    def write_header(self, columns_schema: TTableSchemaColumns) -> None:  # noqa
        pass

    def write_data(self, items: Sequence[TDataItem]) -> None:
        self.items_count += len(items)

    def write_footer(self) -> None:  # noqa
        pass

    def close(self) -> None:  # noqa
        pass

    def write_all(self, columns_schema: TTableSchemaColumns, items: Sequence[TDataItem]) -> None:
        self.write_header(columns_schema)
        self.write_data(items)
        self.write_footer()

    @classmethod
    @abc.abstractmethod
    def writer_spec(cls) -> FileWriterSpec:
        pass

    @classmethod
    def from_file_format(
        cls,
        file_format: TLoaderFileFormat,
        data_item_format: TDataItemFormat,
        f: IO[Any],
        caps: DestinationCapabilitiesContext = None,
    ) -> "DataWriter":
        return cls.class_factory(file_format, data_item_format, ALL_WRITERS)(f, caps)

    @classmethod
    def writer_spec_from_file_format(
        cls, file_format: TLoaderFileFormat, data_item_format: TDataItemFormat
    ) -> FileWriterSpec:
        return cls.class_factory(file_format, data_item_format, ALL_WRITERS).writer_spec()

    @classmethod
    def item_format_from_file_extension(cls, extension: str) -> TDataItemFormat:
        """Simple heuristic to get data item format from file extension"""
        if extension == "typed-jsonl":
            return "object"
        elif extension == "parquet":
            return "arrow"
        elif extension == "model":
            return "model"
        # those files may be imported by normalizer as is
        elif extension in LOADER_FILE_FORMATS:
            return "file"
        else:
            raise ValueError(f"No `data_item_format` associated with file extension: `{extension}`")

    @staticmethod
    def writer_class_from_spec(spec: FileWriterSpec) -> Type["DataWriter"]:
        try:
            return WRITER_SPECS[spec]
        except KeyError:
            if spec.data_item_format == "file":
                return ImportFileWriter
            raise FileSpecNotFound(spec.file_format, spec.data_item_format, spec)

    @staticmethod
    def class_factory(
        file_format: TLoaderFileFormat,
        data_item_format: TDataItemFormat,
        writers: Sequence[Type["DataWriter"]],
    ) -> Type["DataWriter"]:
        for writer in writers:
            spec = writer.writer_spec()
            if spec.file_format == file_format and spec.data_item_format == data_item_format:
                return writer
        raise FileFormatForItemFormatNotFound(file_format, data_item_format)


class ImportFileWriter(DataWriter):
    """May only import files, fails on any open/write operations"""

    def write_header(self, columns_schema: TTableSchemaColumns) -> None:
        raise NotImplementedError(
            "`ImportFileWriter` cannot write any files. You have a bug in your code."
        )

    @classmethod
    def writer_spec(cls) -> FileWriterSpec:
        raise NotImplementedError("`ImportFileWriter` has no single spec")


class JsonlWriter(DataWriter):
    def write_data(self, items: Sequence[TDataItem]) -> None:
        super().write_data(items)
        for row in items:
            json.dump(row, self._f)
            self._f.write(b"\n")

    @classmethod
    def writer_spec(cls) -> FileWriterSpec:
        return FileWriterSpec(
            "jsonl",
            "object",
            file_extension="jsonl",
            is_binary_format=True,
            supports_schema_changes="True",
            supports_compression=True,
        )


class ModelWriter(DataWriter):
    """Writes incoming items row by row into a text file and ensures a trailing ;"""

    def write_header(self, columns_schema: TTableSchemaColumns) -> None:
        pass

    def write_data(self, items: Sequence[TDataItem]) -> None:
        super().write_data(items)
        for item in items:
            dialect = item.query_dialect
            query = item.to_sql()
            self._f.write("dialect: " + (dialect or "") + "\n" + query + "\n")

    @classmethod
    def writer_spec(cls) -> FileWriterSpec:
        return FileWriterSpec(
            "model",
            "model",
            file_extension="model",
            is_binary_format=False,
            supports_schema_changes="True",
            supports_compression=False,
            # NOTE: we create a new model file for each sql row
            file_max_items=1,
        )


class TypedJsonlListWriter(JsonlWriter):
    def write_data(self, items: Sequence[TDataItem]) -> None:
        # skip JsonlWriter when calling super
        super(JsonlWriter, self).write_data(items)
        # write all rows as one list which will require to write just one line
        # encode types with PUA characters
        json.typed_dump(items, self._f)
        self._f.write(b"\n")

    @classmethod
    def writer_spec(cls) -> FileWriterSpec:
        return FileWriterSpec(
            "typed-jsonl",
            "object",
            file_extension="typed-jsonl",
            is_binary_format=True,
            supports_schema_changes="True",
            supports_compression=True,
        )


class InsertValuesWriter(DataWriter):
    def __init__(self, f: IO[Any], caps: DestinationCapabilitiesContext = None) -> None:
        assert (
            caps is not None
        ), "InsertValuesWriter requires destination capabilities to be present"
        super().__init__(f, caps)
        self._chunks_written = 0
        self._headers_lookup: Dict[str, int] = None
        self.writer_type = caps.insert_values_writer_type
        if self.writer_type == "default":
            self.pre, self.post, self.sep = ("(", ")", ",\n")
        elif self.writer_type == "select_union":
            self.pre, self.post, self.sep = ("SELECT ", "", " UNION ALL\n")

    def write_header(self, columns_schema: TTableSchemaColumns) -> None:
        assert self._chunks_written == 0
        assert columns_schema is not None, "column schema required"
        headers = columns_schema.keys()
        # dict lookup is always faster
        self._headers_lookup = {v: i for i, v in enumerate(headers)}
        # do not write INSERT INTO command, this must be added together with table name by the loader
        self._f.write("INSERT INTO {}(")
        self._f.write(",".join(map(self._caps.escape_identifier, headers)))
        self._f.write(")\n")
        if self.writer_type == "default":
            self._f.write("VALUES\n")

    def write_data(self, items: Sequence[TDataItem]) -> None:
        super().write_data(items)

        # do not write empty rows, such things may be produced by Arrow adapters
        if len(items) == 0:
            return

        def write_row(row: StrAny, last_row: bool = False) -> None:
            output = ["NULL"] * len(self._headers_lookup)
            for n, v in row.items():
                output[self._headers_lookup[n]] = self._caps.escape_literal(v)
            self._f.write(self.pre)
            self._f.write(",".join(output))
            self._f.write(self.post)
            if not last_row:
                self._f.write(self.sep)

        # if next chunk add separator
        if self._chunks_written > 0:
            self._f.write(self.sep)

        # write rows
        for row in items[:-1]:
            write_row(row)

        # write last row without separator so we can write footer eventually
        write_row(items[-1], last_row=True)
        self._chunks_written += 1

    def write_footer(self) -> None:
        if self._chunks_written > 0:
            self._f.write(";")

    @classmethod
    def writer_spec(cls) -> FileWriterSpec:
        return FileWriterSpec(
            "insert_values",
            "object",
            file_extension="insert_values",
            is_binary_format=False,
            supports_schema_changes="Buffer",
            supports_compression=True,
            requires_destination_capabilities=True,
        )


class ParquetDataWriter(DataWriter):
    @with_config(spec=ParquetFormatConfiguration)
    def __init__(
        self,
        f: IO[Any],
        caps: DestinationCapabilitiesContext = None,
        *,
        flavor: Optional[str] = None,
        version: Optional[str] = "2.4",
        data_page_size: Optional[int] = None,
        timestamp_timezone: str = "UTC",
        row_group_size: Optional[int] = None,
        coerce_timestamps: Optional[Literal["s", "ms", "us", "ns"]] = None,
        allow_truncated_timestamps: bool = False,
        use_compliant_nested_type: bool = True,
        _format: ParquetFormatConfiguration = None,  # will receive the full config
    ) -> None:
        super().__init__(f, caps or DestinationCapabilitiesContext.generic_capabilities("parquet"))
        from dlt.common.libs.pyarrow import pyarrow

        self.writer: Optional[pyarrow.parquet.ParquetWriter] = None
        self.schema: Optional[pyarrow.Schema] = None
        self.nested_indices: List[str] = None
        # merge parquet format
        if self._caps.parquet_format is not None:
            self.parquet_format = self._caps.parquet_format.copy()
            self.parquet_format.update(_format.as_dict_nondefault())
        else:
            self.parquet_format = _format

    def _create_writer(self, schema: "pa.Schema") -> "pa.parquet.ParquetWriter":
        from dlt.common.libs.pyarrow import pyarrow

        # if timestamps are not explicitly coerced, use destination resolution
        # TODO: introduce maximum timestamp resolution, using timestamp_precision too aggressive
        # if not self.coerce_timestamps:
        #     self.coerce_timestamps = get_py_arrow_timestamp(
        #         self._caps.timestamp_precision, "UTC"
        #     ).unit
        #     self.allow_truncated_timestamps = True

        parquet_writer_kwargs = {
            "flavor": self.parquet_format.flavor,
            "version": self.parquet_format.version,
            "data_page_size": self.parquet_format.data_page_size,
            "coerce_timestamps": self.parquet_format.coerce_timestamps,
            "allow_truncated_timestamps": self.parquet_format.allow_truncated_timestamps,
            "use_compliant_nested_type": self.parquet_format.use_compliant_nested_type,
            "write_page_index": self.parquet_format.write_page_index,
        }
        if Version(pyarrow.__version__).major >= 21:
            parquet_writer_kwargs["use_content_defined_chunking"] = (
                self.parquet_format.use_content_defined_chunking
            )
        return pyarrow.parquet.ParquetWriter(self._f, schema, **parquet_writer_kwargs)

    def write_header(self, columns_schema: TTableSchemaColumns) -> None:
        from dlt.common.libs.pyarrow import columns_to_arrow

        # build schema
        self.schema = columns_to_arrow(
            columns_schema, self._caps, self.parquet_format.timestamp_timezone
        )
        # find row items that are of the json type (could be abstracted out for use in other writers?)
        self.nested_indices = [
            i for i, field in columns_schema.items() if field["data_type"] == "json"
        ]
        self.writer = self._create_writer(self.schema)

    def write_data(self, items: Sequence[TDataItem]) -> None:
        super().write_data(items)
        from dlt.common.libs.pyarrow import pyarrow

        # serialize json types and replace with strings
        for key in self.nested_indices:
            for row in items:
                if (value := row.get(key)) is not None:
                    # TODO: make this configurable
                    if value is not None and not isinstance(value, str):
                        row[key] = json.dumps(value)

        table = pyarrow.Table.from_pylist(items, schema=self.schema)
        # detect non-null columns receiving nulls. above v.19 it is checked in `write_table`
        if Version(pyarrow.__version__).major < 19:
            table = table.cast(self.schema)
        # Write
        self.writer.write_table(table, row_group_size=self.parquet_format.row_group_size)

    def close(self) -> None:  # noqa
        if self.writer:
            self.writer.close()
            self.writer = None

    @classmethod
    def writer_spec(cls) -> FileWriterSpec:
        return FileWriterSpec(
            "parquet",
            "object",
            "parquet",
            is_binary_format=True,
            supports_schema_changes="Buffer",
            requires_destination_capabilities=True,
            supports_compression=False,
        )


class CsvWriter(DataWriter):
    @with_config(spec=CsvFormatConfiguration)
    def __init__(
        self,
        f: IO[Any],
        caps: DestinationCapabilitiesContext = None,
        *,
        delimiter: str = ",",
        include_header: bool = True,
        quoting: CsvQuoting = "quote_needed",
        lineterminator: str = "\n",
        bytes_encoding: str = "utf-8",
    ) -> None:
        super().__init__(f, caps)
        self.include_header = include_header
        self.delimiter = delimiter
        self.quoting: CsvQuoting = quoting
        self.lineterminator = lineterminator
        self.writer: csv.DictWriter[str] = None
        self.bytes_encoding = bytes_encoding

    def write_header(self, columns_schema: TTableSchemaColumns) -> None:
        self._columns_schema = columns_schema
        if self.quoting == "quote_needed":
            quoting: Literal[0, 1, 2, 3] = csv.QUOTE_NONNUMERIC
        elif self.quoting == "quote_all":
            quoting = csv.QUOTE_ALL
        elif self.quoting == "quote_none":
            quoting = csv.QUOTE_NONE
        elif self.quoting == "quote_minimal":
            quoting = csv.QUOTE_MINIMAL
        else:
            raise ValueError(self.quoting)

        self.writer = csv.DictWriter(
            self._f,
            fieldnames=list(columns_schema.keys()),
            extrasaction="ignore",
            dialect=csv.unix_dialect,
            delimiter=self.delimiter,
            quoting=quoting,
            lineterminator=self.lineterminator,
        )
        if self.include_header:
            self.writer.writeheader()
        # find row items that are of the json type
        self.nested_indices = [
            i for i, field in columns_schema.items() if field["data_type"] == "json"
        ]
        # find row items that are of the binary type
        self.bytes_indices = [
            i for i, field in columns_schema.items() if field["data_type"] == "binary"
        ]

    def write_data(self, items: Sequence[TDataItem]) -> None:
        # convert bytes and json
        if self.nested_indices or self.bytes_indices:
            for row in items:
                for key in self.nested_indices:
                    if (value := row.get(key)) is not None:
                        row[key] = json.dumps(value)
                for key in self.bytes_indices:
                    if (value := row.get(key)) is not None:
                        # assumed bytes value
                        try:
                            row[key] = value.decode(self.bytes_encoding)
                        except UnicodeError:
                            raise InvalidDataItem(
                                "csv",
                                "object",
                                f"'{key}' contains bytes that cannot be decoded with encoding"
                                f" `{self.bytes_encoding}`. Remove binary columns or replace their"
                                " content with a hex representation: \\x... while keeping data"
                                " type as binary.",
                            )

        self.writer.writerows(items)
        # count rows that got written
        self.items_count += sum(len(row) for row in items)

    def close(self) -> None:
        self.writer = None
        self._first_schema = None

    @classmethod
    def writer_spec(cls) -> FileWriterSpec:
        return FileWriterSpec(
            "csv",
            "object",
            file_extension="csv",
            is_binary_format=False,
            supports_schema_changes="False",
            requires_destination_capabilities=False,
            supports_compression=True,
        )


class ArrowToParquetWriter(ParquetDataWriter):
    def write_header(self, columns_schema: TTableSchemaColumns) -> None:
        # Schema will be written as-is from the arrow table
        self._column_schema = columns_schema

    def write_data(self, items: Sequence[TDataItem]) -> None:
        from dlt.common.libs.pyarrow import concat_batches_and_tables_in_order

        if not items:
            return
        # concat batches and tables into a single one, preserving order
        # pyarrow writer starts a row group for each item it writes (even with 0 rows)
        # it also converts batches into tables internally. by creating a single table
        # we allow the user rudimentary control over row group size via max buffered items
        table = concat_batches_and_tables_in_order(items)
        # release batch references - concat is zero-copy so table shares the
        # underlying buffers via Arrow refcounting. clearing the input list
        # drops the Python-level RecordBatch/Table references so only the
        # concatenated table keeps the buffers alive
        if isinstance(items, list):
            items.clear()
        self.items_count += table.num_rows
        if not self.writer:
            self.writer = self._create_writer(table.schema)
        # write concatenated tables
        self.writer.write_table(table, row_group_size=self.parquet_format.row_group_size)

    def write_footer(self) -> None:
        if not self.writer:
            raise NotImplementedError("Arrow Writer does not support writing empty files")
        return super().write_footer()

    def close(self) -> None:
        return super().close()

    @classmethod
    def writer_spec(cls) -> FileWriterSpec:
        return FileWriterSpec(
            "parquet",
            "arrow",
            file_extension="parquet",
            is_binary_format=True,
            supports_schema_changes="False",
            requires_destination_capabilities=False,
            supports_compression=False,
        )


class ArrowToCsvWriter(DataWriter):
    @with_config(spec=CsvFormatConfiguration)
    def __init__(
        self,
        f: IO[Any],
        caps: DestinationCapabilitiesContext = None,
        *,
        delimiter: str = ",",
        include_header: bool = True,
        quoting: CsvQuoting = "quote_needed",
    ) -> None:
        super().__init__(f, caps)
        self.delimiter = delimiter
        self._delimiter_b = delimiter.encode("ascii")
        self.include_header = include_header
        self.quoting: CsvQuoting = quoting
        self.writer: Any = None

    def write_header(self, columns_schema: TTableSchemaColumns) -> None:
        self._columns_schema = columns_schema

    def write_data(self, items: Sequence[TDataItem]) -> None:
        from dlt.common.libs.pyarrow import pyarrow
        import pyarrow.csv

        for item in items:
            if isinstance(item, (pyarrow.Table, pyarrow.RecordBatch)):
                if not self.writer:
                    try:
                        self.writer = pyarrow.csv.CSVWriter(
                            self._f,
                            item.schema,
                            write_options=pyarrow.csv.WriteOptions(
                                # set include_header to False to handle header separately until
                                # https://github.com/apache/arrow/issues/47575 is released
                                # see _make_csv_header() for details
                                include_header=False,
                                delimiter=self._delimiter_b,
                                quoting_style=self._get_arrow_quoting_style(),
                            ),
                        )
                        if self.include_header:
                            self._f.write(self._make_csv_header())
                        self._first_schema = item.schema
                    except pyarrow.ArrowInvalid as inv_ex:
                        if "Unsupported Type" in str(inv_ex):
                            raise InvalidDataItem(
                                "csv",
                                "arrow",
                                "Arrow data contains a column that cannot be written to csv file"
                                f" `{inv_ex}`. Remove nested columns (struct, map) or convert them"
                                " to json strings.",
                            )
                        raise
                # make sure that Schema stays the same
                if not item.schema.equals(self._first_schema):
                    raise InvalidDataItem(
                        "csv",
                        "arrow",
                        "Arrow schema changed without rotating the file. This may be internal"
                        " error or misuse of the writer.\nFirst"
                        f" schema:\n{self._first_schema}\n\nCurrent schema:\n{item.schema}",
                    )

                # write headers only on the first write
                try:
                    self.writer.write(item)
                except pyarrow.ArrowInvalid as inv_ex:
                    if "Invalid UTF8 payload" in str(inv_ex):
                        raise InvalidDataItem(
                            "csv",
                            "arrow",
                            "Arrow data contains string or binary columns with invalid UTF-8"
                            " characters. Remove binary columns or replace their content with a hex"
                            " representation: \\x... while keeping data type as binary.",
                        )
                    if "Timezone database not found" in str(inv_ex):
                        raise InvalidDataItem(
                            "csv",
                            "arrow",
                            str(inv_ex)
                            + ". Arrow does not ship with tzdata on Windows. You need to install it"
                            " yourself:"
                            " https://arrow.apache.org/docs/cpp/build_system.html#runtime-dependencies",
                        )
                    raise
            else:
                raise ValueError(f"Unsupported type `{type(item)}`")
            # count rows that got written
            self.items_count += item.num_rows

    def write_footer(self) -> None:
        default_arrow_line_terminator = b"\n"
        if self.writer is None and self.include_header:
            # empty file: emit only the header line (no data rows)
            self._f.write(self._make_csv_header().rstrip(default_arrow_line_terminator))

    def close(self) -> None:
        if self.writer:
            self.writer.close()
            self.writer = None
            self._first_schema = None

    @classmethod
    def writer_spec(cls) -> FileWriterSpec:
        return FileWriterSpec(
            "csv",
            "arrow",
            file_extension="csv",
            is_binary_format=True,
            supports_schema_changes="False",
            requires_destination_capabilities=False,
            supports_compression=True,
        )

    def _get_arrow_quoting_style(self) -> str:
        if self.quoting == "quote_needed":
            return "needed"
        elif self.quoting == "quote_all":
            return "all_valid"
        elif self.quoting == "quote_none":
            return "none"
        else:
            raise ValueError(self.quoting)

    def _make_csv_header(self) -> bytes:
        # In pyarrow 21.0.0, the CSVWriter does not support specifying the header quote style.
        # This is a workaround to create a header which respects the quote style.
        # See https://github.com/apache/arrow/issues/47575 for details.
        # This needs to be removed once https://github.com/apache/arrow/issues/47575 is released.
        from dlt.common.libs.pyarrow import pyarrow
        import pyarrow.csv

        names = [col["name"] for col in self._columns_schema.values()]
        arrays = [pyarrow.array([n]) for n in names]
        schema = pyarrow.schema([pyarrow.field(n, pyarrow.string()) for n in names])
        table = pyarrow.Table.from_arrays(arrays, schema=schema)

        # Write into an in-memory Arrow sink so schema doesn't affect the real writer
        sink = pyarrow.BufferOutputStream()
        header_writer = pyarrow.csv.CSVWriter(
            sink,
            schema,
            write_options=pyarrow.csv.WriteOptions(
                include_header=False,
                delimiter=self._delimiter_b,
                quoting_style=self._get_arrow_quoting_style(),
            ),
        )
        header_writer.write(table)
        header_writer.close()
        return cast(bytes, sink.getvalue().to_pybytes())


class ArrowToObjectAdapter:
    """A mixin that will convert object writer into arrow writer."""

    def write_data(self, items: Sequence[TDataItem]) -> None:
        for batch in items:
            # convert to object data item format
            super().write_data(batch.to_pylist())  # type: ignore[misc]

    @staticmethod
    def convert_spec(base: Type[DataWriter]) -> FileWriterSpec:
        spec = base.writer_spec()
        return spec._replace(data_item_format="arrow")


class ArrowToInsertValuesWriter(ArrowToObjectAdapter, InsertValuesWriter):
    @classmethod
    def writer_spec(cls) -> FileWriterSpec:
        return cls.convert_spec(InsertValuesWriter)


class ArrowToJsonlWriter(ArrowToObjectAdapter, JsonlWriter):
    @classmethod
    def writer_spec(cls) -> FileWriterSpec:
        return cls.convert_spec(JsonlWriter)


class ArrowToTypedJsonlListWriter(ArrowToObjectAdapter, TypedJsonlListWriter):
    @classmethod
    def writer_spec(cls) -> FileWriterSpec:
        return cls.convert_spec(TypedJsonlListWriter)


def is_native_writer(writer_type: Type[DataWriter]) -> bool:
    """Checks if writer has adapter mixin. Writers with adapters are not native and typically
    decrease the performance.
    """
    # we only have arrow adapters now
    return not issubclass(writer_type, ArrowToObjectAdapter)


ALL_WRITERS: List[Type[DataWriter]] = [
    JsonlWriter,
    TypedJsonlListWriter,
    InsertValuesWriter,
    ParquetDataWriter,
    CsvWriter,
    ArrowToParquetWriter,
    ArrowToInsertValuesWriter,
    ArrowToJsonlWriter,
    ArrowToTypedJsonlListWriter,
    ArrowToCsvWriter,
    ModelWriter,
]

WRITER_SPECS: Dict[FileWriterSpec, Type[DataWriter]] = {
    writer.writer_spec(): writer for writer in ALL_WRITERS
}

NATIVE_FORMAT_WRITERS: Dict[TDataItemFormat, Tuple[Type[DataWriter], ...]] = {
    # all "object" writers are native object writers (no adapters yet)
    "object": tuple(
        writer
        for writer in ALL_WRITERS
        if writer.writer_spec().data_item_format == "object" and is_native_writer(writer)
    ),
    # exclude arrow adapters
    "arrow": tuple(
        writer
        for writer in ALL_WRITERS
        if writer.writer_spec().data_item_format == "arrow" and is_native_writer(writer)
    ),
    "model": tuple(
        writer
        for writer in ALL_WRITERS
        if writer.writer_spec().data_item_format == "model" and is_native_writer(writer)
    ),
}


def resolve_best_writer_spec(
    item_format: TDataItemFormat,
    possible_file_formats: Sequence[TLoaderFileFormat],
    preferred_format: TLoaderFileFormat = None,
) -> FileWriterSpec:
    """Finds best writer for `item_format` out of `possible_file_formats`. Tries `preferred_format` first.
    Best possible writer is a native writer for `item_format` writing files in `preferred_format`.
    If not found, any native writer for `possible_file_formats` is picked.
    Native writer supports `item_format` directly without a need to convert to other item formats.
    """
    native_writers = NATIVE_FORMAT_WRITERS[item_format]
    # check if preferred format has native item_format writer
    if preferred_format:
        if preferred_format not in possible_file_formats:
            raise ValueErrorWithKnownValues(
                "preferred_format", preferred_format, possible_file_formats
            )

        try:
            return DataWriter.class_factory(
                preferred_format, item_format, native_writers
            ).writer_spec()
        except DataWriterNotFound:
            pass
    # if not found, use scan native file formats for item format
    for supported_format in possible_file_formats:
        if supported_format != preferred_format:
            try:
                return DataWriter.class_factory(
                    supported_format, item_format, native_writers
                ).writer_spec()
            except DataWriterNotFound:
                pass

    # search all writers
    if preferred_format:
        try:
            return DataWriter.class_factory(
                preferred_format, item_format, ALL_WRITERS
            ).writer_spec()
        except DataWriterNotFound:
            pass

    for supported_format in possible_file_formats:
        if supported_format != preferred_format:
            try:
                return DataWriter.class_factory(
                    supported_format, item_format, ALL_WRITERS
                ).writer_spec()
            except DataWriterNotFound:
                pass

    raise SpecLookupFailed(item_format, possible_file_formats, preferred_format)


def get_best_writer_spec(
    item_format: TDataItemFormat, file_format: TLoaderFileFormat
) -> FileWriterSpec:
    """Gets writer for `item_format` writing files in {file_format}. Looks for native writer first"""
    native_writers = NATIVE_FORMAT_WRITERS[item_format]
    try:
        return DataWriter.class_factory(file_format, item_format, native_writers).writer_spec()
    except DataWriterNotFound:
        return DataWriter.class_factory(file_format, item_format, ALL_WRITERS).writer_spec()


def create_import_spec(
    item_file_format: TLoaderFileFormat,
    possible_file_formats: Sequence[TLoaderFileFormat],
) -> FileWriterSpec:
    """Creates writer spec that may be used only to import files"""
    # can the item file be directly imported?
    if item_file_format not in possible_file_formats:
        raise SpecLookupFailed("file", possible_file_formats, item_file_format)

    spec = DataWriter.class_factory(item_file_format, "object", ALL_WRITERS).writer_spec()
    return spec._replace(data_item_format="file")


def count_rows_in_items(item: TDataItems) -> int:
    """Count total number of rows of `items` which may be
    * single item
    * list of single items
    * list of tables like data frames or arrow
    """

    if isinstance(item, list):
        # if item supports "shape" it will be used to count items
        if len(item) > 0 and hasattr(item[0], "shape"):
            return sum(len(tbl) for tbl in item)
        else:
            return len(item)
    else:
        # update row count, if item supports "num_rows" it will be used to count items
        if hasattr(item, "shape"):
            return len(item)
        else:
            return 1



from pathlib import Path
from typing import Any, ClassVar, Optional, Sequence

from dlt.common.configuration import configspec
from dlt.common.configuration.specs import BaseConfiguration, known_sections
from dlt.common.configuration.specs.pluggable_run_context import PluggableRunContext
from dlt.common.typing import AnyFun, TSecretStrValue
from dlt.common.utils import digest256, main_module_file_path
from dlt.common.destination import TDestinationReferenceArg
from dlt.common.pipeline import TRefreshMode
from dlt.common.configuration.exceptions import ConfigurationValueError


@configspec
class PipelineRuntimeConfiguration(BaseConfiguration):
    slack_incoming_hook: Optional[TSecretStrValue] = None
    # mcp_config: McpConfiguration
    pluggable_run_context: PluggableRunContext = None
    """Pluggable run context with current run context"""
    intercept_signals: bool = True
    """Enable/disable signal handlers in normalize and load steps"""

    __section__: ClassVar[str] = "runtime"


@configspec
class PipelineConfiguration(BaseConfiguration):
    pipeline_name: Optional[str] = None
    pipelines_dir: Optional[str] = None

    destination_type: Optional[TDestinationReferenceArg] = None
    destination_name: Optional[str] = None

    staging_type: Optional[TDestinationReferenceArg] = None
    staging_name: Optional[str] = None

    import_schema_path: Optional[str] = None
    export_schema_path: Optional[str] = None

    dataset_name: Optional[str] = None
    dataset_name_layout: Optional[str] = None
    """Layout for dataset_name, where %s is replaced with dataset_name. For example: 'prefix_%s'"""
    pipeline_salt: Optional[TSecretStrValue] = None
    restore_from_destination: bool = True
    """Enables the `run` method of the `Pipeline` object to restore the pipeline state and schemas from the destination"""
    enable_runtime_trace: bool = True
    """Enables the tracing. Tracing saves the execution trace locally and is required by `dlt deploy`."""
    use_single_dataset: bool = True
    """Stores all schemas in single dataset. When False, each schema will get a separate dataset with `{dataset_name}_{schema_name}"""
    full_refresh: Optional[bool] = None
    """Deprecated. Use `dev_mode` instead. When set to True, each instance of the pipeline with the `pipeline_name` starts from scratch when run and loads the data to a separate dataset."""
    dev_mode: bool = False
    """When set to True, each instance of the pipeline with the `pipeline_name` starts from scratch when run and loads the data to a separate dataset."""
    progress: Optional[str] = None
    runtime: PipelineRuntimeConfiguration = None
    refresh: Optional[TRefreshMode] = None
    """Refresh mode for the pipeline to fully or partially reset a source during run. See docstring of `dlt.pipeline` for more details."""

    def on_resolved(self) -> None:
        # generate pipeline name from the entry point script name
        if not self.pipeline_name:
            self.pipeline_name = get_default_pipeline_name(main_module_file_path())
        if not self.pipeline_salt:
            self.pipeline_salt = digest256(self.pipeline_name)
        if self.dataset_name_layout and "%s" not in self.dataset_name_layout:
            raise ConfigurationValueError(
                "The dataset_name_layout must contain a '%s' placeholder for dataset_name. For"
                " example: 'prefix_%s'"
            )

    # recommended layout is pipelines.<name>
    __recommended_sections__: ClassVar[Sequence[str]] = (known_sections.PIPELINES, "")


def ensure_correct_pipeline_kwargs(f: AnyFun, **kwargs: Any) -> None:
    for arg_name in kwargs:
        if not hasattr(PipelineConfiguration, arg_name) and not arg_name.startswith("_dlt"):
            raise TypeError(f"`{f.__name__}` got an unexpected keyword argument `{arg_name}`")


def get_default_pipeline_name(entry_point_file: str) -> str:
    """Generates default pipeline name based on an entry point of the current Python script
    prefixed with "dlt_"
    """
    if entry_point_file:
        entry_point_file = Path(entry_point_file).stem
    return "dlt_" + (entry_point_file or "pipeline")



# flake8: noqa
import humanize
from typing import Any
import os

import dlt
from dlt.common import pendulum
from dlt.sources.credentials import ConnectionStringCredentials

from dlt.sources.sql_database import sql_database, sql_table, Table

import sqlalchemy as sa


def load_select_tables_from_database() -> None:
    """Use the sql_database source to reflect an entire database schema and load select tables from it.

    This example sources data from the public Rfam MySQL database.
    """
    # Create a pipeline
    pipeline = dlt.pipeline(pipeline_name="rfam", destination="duckdb", dataset_name="rfam_data")

    # Credentials for the sample database.
    # Note: It is recommended to configure credentials in `.dlt/secrets.toml` under `sources.sql_database.credentials`
    credentials = ConnectionStringCredentials(
        "mysql+pymysql://rfamro@mysql-rfam-public.ebi.ac.uk:4497/Rfam"
    )
    # To pass the credentials from `secrets.toml`, comment out the above credentials.
    # And the credentials will be automatically read from `secrets.toml`.

    # Configure the source to load a few select tables incrementally
    source_1 = sql_database(credentials).with_resources("family", "clan")

    # Add incremental config to the resources. "updated" is a timestamp column in these tables that gets used as a cursor
    source_1.family.apply_hints(incremental=dlt.sources.incremental("updated"))
    source_1.clan.apply_hints(incremental=dlt.sources.incremental("updated"))

    # Run the pipeline. The merge write disposition merges existing rows in the destination by primary key
    info = pipeline.run(source_1, write_disposition="merge")
    print(info)

    # Load some other tables with replace write disposition. This overwrites the existing tables in destination
    source_2 = sql_database(credentials).with_resources("features", "author")
    info = pipeline.run(source_2, write_disposition="replace")
    print(info)

    # Load a table incrementally with append write disposition
    # this is good when a table only has new rows inserted, but not updated
    source_3 = sql_database(credentials).with_resources("genome")
    source_3.genome.apply_hints(incremental=dlt.sources.incremental("created"))

    info = pipeline.run(source_3, write_disposition="append")
    print(info)


def load_entire_database() -> None:
    """Use the sql_database source to completely load all tables in a database"""
    pipeline = dlt.pipeline(pipeline_name="rfam", destination="duckdb", dataset_name="rfam_data")

    # By default the sql_database source reflects all tables in the schema
    # The database credentials are sourced from the `.dlt/secrets.toml` configuration
    source = sql_database()

    # Run the pipeline. For a large db this may take a while
    info = pipeline.run(source, write_disposition="replace")
    print(humanize.precisedelta(pipeline.last_trace.finished_at - pipeline.last_trace.started_at))
    print(info)


def load_standalone_table_resource() -> None:
    """Load a few known tables with the standalone sql_table resource, request full schema and deferred
    table reflection"""
    pipeline = dlt.pipeline(
        pipeline_name="rfam_database",
        destination="duckdb",
        dataset_name="rfam_data",
        dev_mode=True,
    )

    # Load a table incrementally starting at a given date
    # Adding incremental via argument like this makes extraction more efficient
    # as only rows newer than the start date are fetched from the table
    # we also use `detect_precision_hints` to get detailed column schema
    # and defer_table_reflect to reflect schema only during execution
    family = sql_table(
        credentials=ConnectionStringCredentials(
            "mysql+pymysql://rfamro@mysql-rfam-public.ebi.ac.uk:4497/Rfam"
        ),
        table="family",
        incremental=dlt.sources.incremental(
            "updated",
        ),
        reflection_level="full_with_precision",
        defer_table_reflect=True,
    )
    # columns will be empty here due to defer_table_reflect set to True
    print(family.compute_table_schema())

    # Load all data from another table
    genome = sql_table(
        credentials="mysql+pymysql://rfamro@mysql-rfam-public.ebi.ac.uk:4497/Rfam",
        table="genome",
        reflection_level="full_with_precision",
        defer_table_reflect=True,
    )

    # Run the resources together (just take one page of results to make it faster)
    info = pipeline.extract([family.add_limit(1), genome.add_limit(1)], write_disposition="merge")
    print(info)
    # Show inferred columns
    print(pipeline.default_schema.to_pretty_yaml())


def select_with_end_value_and_row_order() -> None:
    """Gets data from a table withing a specified range and sorts rows descending"""
    pipeline = dlt.pipeline(
        pipeline_name="rfam_database",
        destination="duckdb",
        dataset_name="rfam_data",
        dev_mode=True,
    )

    # gets data from this range
    start_date = pendulum.now().subtract(years=1)
    end_date = pendulum.now()

    family = sql_table(
        credentials="mysql+pymysql://rfamro@mysql-rfam-public.ebi.ac.uk:4497/Rfam",
        table="family",
        incremental=dlt.sources.incremental(  # declares desc row order
            "updated", initial_value=start_date, end_value=end_date, row_order="desc"
        ),
        chunk_size=10,
    )
    # also we do not want the whole table, so we add limit to get just one chunk (10 records)
    pipeline.run(family.add_limit(1))
    # only 10 rows
    print(pipeline.last_trace.last_normalize_info)


def my_sql_via_pyarrow() -> None:
    """Uses pyarrow backend to load tables from mysql"""

    # uncomment line below to get load_id into your data (slows pyarrow loading down)
    # dlt.config["normalize.parquet_normalizer.add_dlt_load_id"] = True

    # Create a pipeline
    pipeline = dlt.pipeline(
        pipeline_name="rfam_cx",
        destination="duckdb",
        dataset_name="rfam_data_arrow_4",
    )

    def _double_as_decimal_adapter(table: sa.Table) -> sa.Table:
        """Return double as double, not decimals, only works if you are using sqlalchemy 2.0"""
        for column in table.columns.values():
            if hasattr(sa, "Double") and isinstance(column.type, sa.Double):
                column.type.asdecimal = False
        return table

    sql_alchemy_source = sql_database(
        "mysql+pymysql://rfamro@mysql-rfam-public.ebi.ac.uk:4497/Rfam?&binary_prefix=true",
        backend="pyarrow",
        table_adapter_callback=_double_as_decimal_adapter,
    ).with_resources("family", "genome")

    info = pipeline.run(sql_alchemy_source)
    print(info)


def create_unsw_flow() -> None:
    """Uploads UNSW_Flow dataset to postgres via csv stream skipping dlt normalizer.
    You need to download the dataset from https://github.com/rdpahalavan/nids-datasets
    """
    from pyarrow.parquet import ParquetFile

    # from dlt.destinations import postgres

    # use those config to get 3x speedup on parallelism
    # [sources.data_writer]
    # file_max_bytes=3000000
    # buffer_max_items=200000

    # [normalize]
    # workers=3

    data_iter = ParquetFile("UNSW-NB15/Network-Flows/UNSW_Flow.parquet").iter_batches(
        batch_size=128 * 1024
    )

    pipeline = dlt.pipeline(
        pipeline_name="unsw_upload",
        # destination=postgres("postgres://loader:loader@localhost:5432/dlt_data"),
        destination="postgres",
        progress="log",
    )
    pipeline.run(
        data_iter,
        dataset_name="speed_test",
        table_name="unsw_flow_7",
        loader_file_format="csv",
    )


def test_connectorx_speed() -> None:
    """Uses unsw_flow dataset (~2mln rows, 25+ columns) to test connectorx speed"""
    import os

    unsw_table = sql_table(
        "postgresql://loader:loader@localhost:5432/dlt_data",
        "unsw_flow_7",
        "speed_test",
        # this is ignored by connectorx
        chunk_size=100000,
        backend="connectorx",
        # keep source data types
        reflection_level="full_with_precision",
        # just to demonstrate how to setup a separate connection string for connectorx
        backend_kwargs={"conn": "postgresql://loader:loader@localhost:5432/dlt_data"},
    )

    pipeline = dlt.pipeline(
        pipeline_name="unsw_download",
        destination="filesystem",
        # destination=filesystem(os.path.abspath("../_storage/unsw")),
        progress="log",
        dev_mode=True,
    )

    info = pipeline.run(
        unsw_table,
        dataset_name="speed_test",
        table_name="unsw_flow",
        loader_file_format="parquet",
    )
    print(info)


def test_pandas_backend_verbatim_decimals() -> None:
    pipeline = dlt.pipeline(
        pipeline_name="rfam_cx",
        destination="duckdb",
        dataset_name="rfam_data_pandas_2",
    )

    def _double_as_decimal_adapter(table: sa.Table) -> sa.Table:
        """Emits decimals instead of floats."""
        for column in table.columns.values():
            if isinstance(column.type, sa.Float):
                column.type.asdecimal = True
        return table

    sql_alchemy_source = sql_database(
        "mysql+pymysql://rfamro@mysql-rfam-public.ebi.ac.uk:4497/Rfam?&binary_prefix=true",
        backend="pandas",
        table_adapter_callback=_double_as_decimal_adapter,
        chunk_size=100000,
        # set coerce_float to False to represent them as string
        backend_kwargs={"coerce_float": False, "dtype_backend": "numpy_nullable"},
        # preserve full typing info. this will parse
        reflection_level="full_with_precision",
    ).with_resources("family", "genome")

    info = pipeline.run(sql_alchemy_source)
    print(info)


def use_type_adapter() -> None:
    """Example use of type adapter to coerce unknown data types"""
    pipeline = dlt.pipeline(
        pipeline_name="dummy",
        destination="postgres",
        dataset_name="dummy",
    )

    def type_adapter(sql_type: Any) -> Any:
        if isinstance(sql_type, sa.ARRAY):
            return sa.JSON()  # Load arrays as JSON
        return sql_type

    sql_alchemy_source = sql_database(
        "postgresql://loader:loader@localhost:5432/dlt_data",
        backend="pyarrow",
        type_adapter_callback=type_adapter,
        reflection_level="full_with_precision",
    ).with_resources("table_with_array_column")

    info = pipeline.run(sql_alchemy_source)
    print(info)


def specify_columns_to_load() -> None:
    """Run the SQL database source with a subset of table columns loaded"""
    pipeline = dlt.pipeline(
        pipeline_name="dummy",
        destination="duckdb",
    )

    # Columns can be specified per table in env var (json array) or in `.dlt/config.toml`
    os.environ["SOURCES__SQL_DATABASE__FAMILY__INCLUDED_COLUMNS"] = '["rfam_acc", "description"]'

    sql_alchemy_source = sql_database(
        "mysql+pymysql://rfamro@mysql-rfam-public.ebi.ac.uk:4497/Rfam?&binary_prefix=true",
        backend="pyarrow",
        reflection_level="full_with_precision",
    ).with_resources("family", "genome")

    info = pipeline.run(sql_alchemy_source)
    print(info)


if __name__ == "__main__":
    # Load selected tables with different settings
    # load_select_tables_from_database()

    # load_entire_database()
    # select_with_end_value_and_row_order()

    # Load tables with the standalone table resource
    load_standalone_table_resource()

    # Load all tables from the database.
    # Warning: The sample database is very large
    # load_entire_database()



import dataclasses

import os
from typing import Dict, Final, Optional, Type

from dlt.common.typing import DictStrAny, DictStrOptionalStr

from dlt.common import logger
from dlt.common.configuration import configspec, resolve_type
from dlt.common.configuration.specs.hf_credentials import HfCredentials
from dlt.common.destination.client import (
    CredentialsConfiguration,
    DestinationClientStagingConfiguration,
)
from dlt.common.storages import FilesystemConfigurationWithLocalFiles

from dlt.destinations.impl.filesystem.typing import TCurrentDateTime, TExtraPlaceholders
from dlt.destinations.path_utils import check_layout, get_unused_placeholders


@configspec
class FilesystemDestinationClientConfiguration(FilesystemConfigurationWithLocalFiles, DestinationClientStagingConfiguration):  # type: ignore[misc]
    destination_type: Final[str] = dataclasses.field(  # type: ignore[misc]
        default="filesystem", init=False, repr=False, compare=False
    )
    current_datetime: Optional[TCurrentDateTime] = None
    extra_placeholders: Optional[TExtraPlaceholders] = None
    max_state_files: int = 100
    """Maximum number of pipeline state files to keep; 0 or negative value disables cleanup."""
    always_refresh_views: bool = False
    """Always refresh table scanner views by setting the newest table metadata or globbing table files"""
    deltalake_storage_options: Optional[DictStrAny] = None
    """Additional storage options passed to `deltalake` library, overriding credentials-derived values."""
    deltalake_configuration: Optional[DictStrOptionalStr] = None
    """Delta table configuration passed to `write_deltalake` and `create_deltalake` calls."""
    deltalake_streamed_exec: bool = True
    """When true, delta merge operations use streamed execution to reduce memory usage."""
    iceberg_table_properties: Optional[Dict[str, str]] = None
    """Default Iceberg table properties applied to all tables; per-table adapter properties take precedence."""
    iceberg_namespace_properties: Optional[Dict[str, str]] = None
    """Properties passed to the Iceberg catalog when creating the namespace."""

    @resolve_type("credentials")
    def resolve_credentials_type(self) -> Type[CredentialsConfiguration]:
        return super().resolve_credentials_type()

    def on_resolved(self) -> None:
        # Validate layout and show unused placeholders
        _, layout_placeholders = check_layout(self.layout, self.extra_placeholders)
        unused_placeholders = get_unused_placeholders(
            layout_placeholders, list((self.extra_placeholders or {}).keys())
        )
        if unused_placeholders:
            logger.info(f"Found unused layout placeholders: {', '.join(unused_placeholders)}")


@configspec
class HfFilesystemDestinationClientConfiguration(FilesystemDestinationClientConfiguration):
    credentials: HfCredentials = None
    hf_dataset_card: bool = True
    """Create and update dataset card (README.md) with subset metadata for the Hugging Face dataset viewer."""

    @property
    def hf_namespace(self) -> str:
        return os.path.basename(self.bucket_url)



from abc import ABC, abstractmethod
from enum import IntEnum
from pathlib import Path
from typing import Any, ClassVar, Dict, List, Literal, NamedTuple, Optional, Tuple, Type, Union

from dlt.common.runtime.exec_info import is_claude_code, is_codex, is_cursor

from dlt._workspace.cli.ai.utils import (
    ensure_cursor_rule_frontmatter,
    home_dir,
    merge_json_mcp_servers,
    merge_toml_mcp_servers,
    parse_json_mcp,
    parse_toml_mcp,
    read_agents_md_template,
    strip_rule_frontmatter,
    wrap_as_skill,
)
from dlt._workspace.cli.formatters import merge_agents_md_skills

TComponentType = Literal["skill", "command", "rule", "ignore", "mcp"]
TInstallOp = Literal["copytree", "save"]


class AgentDetectLevel(IntEnum):
    ENV = 0
    LOCAL = 1
    GLOBAL = 2


class InstallAction(NamedTuple):
    kind: TComponentType
    source_name: str
    dest_path: Path
    op: TInstallOp
    content_or_path: Union[str, Path]
    conflict: bool
    source_kind: Optional[TComponentType] = None
    skip_index: bool = False


class _AIAgent(ABC):
    """AI coding agent adapter for toolkit installation, detection, and MCP configuration."""

    @property
    @abstractmethod
    def name(self) -> str: ...

    _DIRS: ClassVar[Dict[TComponentType, str]]
    """maps component types to relative directory paths under project root"""

    _GLOBAL_MARKER: ClassVar[str]
    """home-dir marker that proves the tool is installed (e.g. ".claude")"""

    _LOCAL_PROBES: ClassVar[Tuple[str, ...]]
    """project-root paths that indicate usage of this agent"""

    _RULE_EXT: ClassVar[str] = ".md"
    """file extension for rule files"""

    @property
    @abstractmethod
    def ignore_file_name(self) -> str:
        """Platform-specific ignore file name (e.g. .claudeignore)."""

    def component_dir(self, component_type: TComponentType, project_root: Path) -> Path:
        """Root directory for the given component type."""
        if component_type == "ignore":
            return project_root
        assert component_type in self._DIRS, "%s not in _DIRS for %s" % (component_type, self.name)
        return project_root / self._DIRS[component_type]

    def install_actions(
        self,
        component_type: TComponentType,
        content_or_path: Union[str, Path],
        source_name: str,
        toolkit_name: str,
        project_root: Path,
        overwrite: bool = False,
    ) -> List["InstallAction"]:
        """Build install actions for a single component.

        Handles all component types: skill (copytree), ignore (save), and
        delegates command/rule to _install_command_or_rule().

        Args:
            component_type: "skill", "command", "rule", or "ignore"
            content_or_path: file content (str) or source directory (Path)
            source_name: original name (stem or dir name)
            toolkit_name: toolkit identifier
            project_root: target project root
            overwrite: if True, ignore existing conflicts

        Returns:
            list of InstallAction to execute
        """
        if component_type == "skill":
            assert isinstance(content_or_path, Path)
            dest = self.component_dir("skill", project_root) / source_name
            return [
                InstallAction(
                    kind="skill",
                    source_name=source_name,
                    dest_path=dest,
                    op="copytree",
                    content_or_path=content_or_path,
                    conflict=not overwrite and dest.exists(),
                )
            ]
        if component_type == "ignore":
            assert isinstance(content_or_path, str)
            dest = self.component_dir("ignore", project_root) / self.ignore_file_name
            return [
                InstallAction(
                    kind="ignore",
                    source_name=".claudeignore",
                    dest_path=dest,
                    op="save",
                    content_or_path=content_or_path,
                    conflict=not overwrite and dest.exists(),
                )
            ]
        assert isinstance(content_or_path, str)
        return self._install_command_or_rule(
            component_type, content_or_path, source_name, toolkit_name, project_root, overwrite
        )

    def _transform_rule(self, content: str) -> str:
        """Transform rule content before writing. Override for agent-specific formatting."""
        return content

    def _install_command_or_rule(
        self,
        component_type: TComponentType,
        content: str,
        source_name: str,
        toolkit_name: str,
        project_root: Path,
        overwrite: bool,
    ) -> List["InstallAction"]:
        """Install actions for commands and rules.

        Commands are saved as `source_name.md` under the command dir.
        Rules are transformed via `_transform_rule` and saved as
        `toolkit_name-source_name{_RULE_EXT}` under the rule dir.
        """
        if component_type == "command":
            dest = self.component_dir("command", project_root) / (source_name + ".md")
            return [
                InstallAction(
                    kind="command",
                    source_name=source_name,
                    dest_path=dest,
                    op="save",
                    content_or_path=content,
                    conflict=not overwrite and dest.exists(),
                )
            ]
        transformed = self._transform_rule(content)
        out_filename = toolkit_name + "-" + source_name + self._RULE_EXT
        dest = self.component_dir("rule", project_root) / out_filename
        return [
            InstallAction(
                kind="rule",
                source_name=source_name,
                dest_path=dest,
                op="save",
                content_or_path=transformed,
                conflict=not overwrite and dest.exists(),
            )
        ]

    def finalize_actions(
        self,
        actions: List["InstallAction"],
        project_root: Path,
        workbench_base: Optional[Path] = None,
    ) -> List["InstallAction"]:
        """Post-process the full action list. Default is identity.

        Subclasses override to coalesce shared targets (e.g. AGENTS.md).
        *workbench_base* is the root of the fetched workbench repo.
        """
        return actions

    @abstractmethod
    def mcp_config_path(self, project_root: Path) -> Path:
        """Platform-specific MCP config file path."""

    @abstractmethod
    def parse_mcp_servers(self, content: str) -> Dict[str, Any]:
        """Parse existing MCP config content into {server_name: config}."""

    @abstractmethod
    def merge_mcp_servers(self, existing_content: str, new_servers: Dict[str, Any]) -> str:
        """Merge new server entries into existing config content."""

    @classmethod
    @abstractmethod
    def _is_env(cls) -> bool:
        """Return True when the agent's runtime env var is set."""

    @classmethod
    def _detect(cls, project_root: Path) -> Optional[AgentDetectLevel]:
        """Detect agent via ENV var, local project probes, or global marker.

        LOCAL probes only fire when the global marker (`~/_GLOBAL_MARKER`)
        also exists, preventing false positives from stale project files.
        """
        if cls._is_env():
            return AgentDetectLevel.ENV
        home = home_dir()
        installed = home is not None and (home / cls._GLOBAL_MARKER).exists()
        if installed and any((project_root / p).exists() for p in cls._LOCAL_PROBES):
            return AgentDetectLevel.LOCAL
        if installed:
            return AgentDetectLevel.GLOBAL
        return None

    @classmethod
    def detect_all(cls, project_root: Path) -> List[Tuple["_AIAgent", "AgentDetectLevel"]]:
        """Return all detected AI coding agents sorted by detection level."""
        detected: List[Tuple[AgentDetectLevel, Type[_AIAgent]]] = []
        for variant_cls in AI_AGENTS.values():
            level = variant_cls._detect(project_root)
            if level is not None:
                detected.append((level, variant_cls))
        detected.sort(key=lambda t: t[0])
        return [(variant_cls(), level) for level, variant_cls in detected]


class _ClaudeAgent(_AIAgent):
    _DIRS: ClassVar[Dict[TComponentType, str]] = {
        "skill": ".claude/skills",
        "command": ".claude/commands",
        "rule": ".claude/rules",
    }
    _GLOBAL_MARKER: ClassVar[str] = ".claude"
    _LOCAL_PROBES: ClassVar[Tuple[str, ...]] = (".claude", "CLAUDE.md")

    @property
    def name(self) -> str:
        return "claude"

    @classmethod
    def _is_env(cls) -> bool:
        return is_claude_code()

    @property
    def ignore_file_name(self) -> str:
        return ".claudeignore"

    def _transform_rule(self, content: str) -> str:
        return strip_rule_frontmatter(content)

    def mcp_config_path(self, project_root: Path) -> Path:
        return project_root / ".mcp.json"

    def parse_mcp_servers(self, content: str) -> Dict[str, Any]:
        return parse_json_mcp(content, "mcpServers")[1]

    def merge_mcp_servers(self, existing_content: str, new_servers: Dict[str, Any]) -> str:
        return merge_json_mcp_servers(existing_content, new_servers, "mcpServers", strip_type=False)


class _CursorAgent(_AIAgent):
    _DIRS: ClassVar[Dict[TComponentType, str]] = {
        "skill": ".cursor/skills",
        "command": ".cursor/commands",
        "rule": ".cursor/rules",
    }
    _GLOBAL_MARKER: ClassVar[str] = ".cursor"
    _LOCAL_PROBES: ClassVar[Tuple[str, ...]] = (".cursor", ".cursorignore", ".cursorrules")

    @property
    def name(self) -> str:
        return "cursor"

    @classmethod
    def _is_env(cls) -> bool:
        return is_cursor()

    _RULE_EXT: ClassVar[str] = ".mdc"

    @property
    def ignore_file_name(self) -> str:
        return ".cursorignore"

    def _transform_rule(self, content: str) -> str:
        return ensure_cursor_rule_frontmatter(content)

    def mcp_config_path(self, project_root: Path) -> Path:
        return project_root / ".cursor" / "mcp.json"

    def parse_mcp_servers(self, content: str) -> Dict[str, Any]:
        return parse_json_mcp(content, "mcpServers")[1]

    def merge_mcp_servers(self, existing_content: str, new_servers: Dict[str, Any]) -> str:
        return merge_json_mcp_servers(existing_content, new_servers, "mcpServers", strip_type=True)


class _CodexAgent(_AIAgent):
    _DIRS: ClassVar[Dict[TComponentType, str]] = {
        "skill": ".agents/skills",
    }
    _GLOBAL_MARKER: ClassVar[str] = ".codex"
    _LOCAL_PROBES: ClassVar[Tuple[str, ...]] = (".agents", "AGENTS.md")

    @property
    def name(self) -> str:
        return "codex"

    @classmethod
    def _is_env(cls) -> bool:
        return is_codex()

    @property
    def ignore_file_name(self) -> str:
        return ".codexignore"

    def agents_md_path(self, project_root: Path) -> Path:
        """Path to the AGENTS.md file for skill registration."""
        return project_root / "AGENTS.md"

    def _install_command_or_rule(
        self,
        component_type: TComponentType,
        content: str,
        source_name: str,
        toolkit_name: str,
        project_root: Path,
        overwrite: bool,
    ) -> List["InstallAction"]:
        if component_type == "command":
            wrapped = wrap_as_skill(content, source_name)
            dest = self.component_dir("skill", project_root) / source_name / "SKILL.md"
            return [
                InstallAction(
                    kind="skill",
                    source_name=source_name,
                    dest_path=dest,
                    op="save",
                    content_or_path=wrapped,
                    conflict=not overwrite and dest.exists(),
                    source_kind="command",
                )
            ]
        # rule → always-apply skill (AGENTS.md is handled by finalize_actions)
        skill_name = toolkit_name + "-" + source_name
        wrapped = wrap_as_skill(content, skill_name, always_apply=True)
        skill_dest = self.component_dir("skill", project_root) / skill_name / "SKILL.md"
        return [
            InstallAction(
                kind="skill",
                source_name=source_name,
                dest_path=skill_dest,
                op="save",
                content_or_path=wrapped,
                conflict=not overwrite and skill_dest.exists(),
                source_kind="rule",
            )
        ]

    def finalize_actions(
        self,
        actions: List["InstallAction"],
        project_root: Path,
        workbench_base: Optional[Path] = None,
    ) -> List["InstallAction"]:
        """Coalesce rule-converted-to-skill actions into a single AGENTS.md merge action."""
        skill_names: List[str] = []
        for a in actions:
            if a.source_kind == "rule" and not a.conflict:
                skill_names.append(a.dest_path.parent.name)

        if not skill_names:
            return actions

        agents_md = self.agents_md_path(project_root)
        existing = agents_md.read_text(encoding="utf-8") if agents_md.is_file() else ""
        template = read_agents_md_template(workbench_base)
        merged = merge_agents_md_skills(existing, skill_names, template=template)
        if merged != existing:
            actions.append(
                InstallAction(
                    kind="rule",
                    source_name="AGENTS.md",
                    dest_path=agents_md,
                    op="save",
                    content_or_path=merged,
                    conflict=False,
                    skip_index=True,
                )
            )
        return actions

    def mcp_config_path(self, project_root: Path) -> Path:
        return project_root / ".codex" / "config.toml"

    def parse_mcp_servers(self, content: str) -> Dict[str, Any]:
        return parse_toml_mcp(content, "mcp_servers")[1]

    def merge_mcp_servers(self, existing_content: str, new_servers: Dict[str, Any]) -> str:
        return merge_toml_mcp_servers(existing_content, new_servers, "mcp_servers")


AI_AGENTS: Dict[str, Type[_AIAgent]] = {
    "claude": _ClaudeAgent,
    "cursor": _CursorAgent,
    "codex": _CodexAgent,
}



from typing import Iterable, cast, Any, List

import gzip
from abc import ABC, abstractmethod
from fsspec import AbstractFileSystem

from dlt.common.schema import Schema


class FSClientBase(ABC):
    fs_client: AbstractFileSystem
    schema: Schema

    @property
    @abstractmethod
    def dataset_path(self) -> str:
        """A path within a bucket to tables in a dataset, ending with separator"""
        pass

    @abstractmethod
    def get_table_dir(self, table_name: str) -> str:
        """Returns a directory containing table files, ending with separator.
        Native filesystem paths are used for local filesystems.
        Note that many tables can share the same table dir.
        """
        pass

    @abstractmethod
    def get_table_dirs(self, table_names: Iterable[str]) -> List[str]:
        """returns directories for given table"""
        pass

    @abstractmethod
    def list_table_files(self, table_name: str) -> List[str]:
        """Returns all filepaths for a given table. Native filesystem paths are used for local filesystems."""
        pass

    @abstractmethod
    def truncate_tables(self, table_names: List[str]) -> None:
        """truncates the given table"""
        pass

    def read_bytes(self, path: str, start: Any = None, end: Any = None, **kwargs: Any) -> bytes:
        """reads given file to bytes object"""
        return cast(bytes, self.fs_client.read_bytes(path, start, end, **kwargs))

    def read_text(
        self,
        path: str,
        encoding: Any = "utf-8",
        errors: Any = None,
        newline: Any = None,
        compression: str = None,
        **kwargs: Any,
    ) -> str:
        """reads given file into string, tries gzip and pure text"""
        if compression is None:
            try:
                return self.read_text(path, encoding, errors, newline, "gzip", **kwargs)
            except (gzip.BadGzipFile, OSError):
                pass
        with self.fs_client.open(
            path, mode="rt", compression=compression, encoding=encoding, newline=newline
        ) as f:
            return cast(str, f.read())



import re
import inspect
import sys
from typing import Dict, Tuple, Type, Any, Optional
from inspect import Signature, Parameter

from dlt.common.typing import (
    AnyType,
    AnyFun,
    ConfigValueSentinel,
    NoneType,
    TSecretValue,
    Annotated,
    SecretSentinel,
    get_type_globals,
    resolve_single_annotation,
)
from dlt.common.configuration import configspec, is_valid_hint, is_secret_hint
from dlt.common.configuration.specs import BaseConfiguration
from dlt.common.utils import get_callable_name

# [^.^_]+ splits by . or _
_SLEEPING_CAT_SPLIT = re.compile("[^.^_]+")


def get_spec_name_from_f(f: AnyFun, kind: str = "Configuration") -> str:
    func_name = get_callable_name(f, "__qualname__").replace(
        "<locals>.", ""
    )  # func qual name contains position in the module, separated by dots

    def _first_up(s: str) -> str:
        return s[0].upper() + s[1:]

    return "".join(map(_first_up, _SLEEPING_CAT_SPLIT.findall(func_name))) + kind


def spec_from_signature(
    f: AnyFun,
    sig: Signature,
    include_defaults: bool = True,
    base: Type[BaseConfiguration] = BaseConfiguration,
) -> Tuple[Type[BaseConfiguration], Dict[str, Any]]:
    """Creates a SPEC on base `base1 for a function `f` with signature `sig`.

    All the arguments in `sig` that are valid SPEC hints and have defaults will be part of the SPEC.
    Special default markers for required SPEC fields `dlt.secrets.value` and `dlt.config.value` are sentinel
    string values with a type set to Any during typechecking. The sentinels are defined in dlt.common.typing module.

    The name of a SPEC type is inferred from qualname of `f` and type will refer to `f` module and is unique
    for a module. NOTE: the SPECS are cached in the module by using name as an id.

    Return value is a tuple of SPEC and SPEC fields created from a `sig`.
    """
    name = get_spec_name_from_f(f)
    module = inspect.getmodule(f)
    base_fields = base.get_resolvable_fields()

    # synthesize configuration from the signature
    new_fields: Dict[str, Any] = {}
    sig_base_fields: Dict[str, Any] = {}
    annotations: Dict[str, Any] = {}
    globalns = get_type_globals(f)

    for p in sig.parameters.values():
        # skip *args and **kwargs, skip typical method params
        if p.kind not in (Parameter.VAR_KEYWORD, Parameter.VAR_POSITIONAL) and p.name not in [
            "self",
            "cls",
        ]:
            field_type = (
                AnyType
                if p.annotation == Parameter.empty
                else resolve_single_annotation(p.annotation, globalns=globalns)
            )
            # keep the base fields if sig not annotated
            if (
                p.name in base_fields
                and field_type is AnyType
                and isinstance(p.default, (NoneType, ConfigValueSentinel))
            ):
                sig_base_fields[p.name] = base_fields[p.name]
                continue
            # only valid hints and parameters with defaults are eligible
            if is_valid_hint(field_type) and p.default != Parameter.empty:
                type_from_literal: AnyType = None
                # make type optional if explicit None is provided as default
                if p.default is None:
                    # optional type
                    field_type = Optional[field_type]
                elif isinstance(p.default, ConfigValueSentinel):
                    # check if the defaults were attributes of the form .config.value or .secrets.value
                    type_from_literal = p.default.default_type
                    if type_from_literal is TSecretValue:
                        # override type with secret value if secrets.value
                        if not is_secret_hint(field_type):
                            if field_type is AnyType:
                                field_type = TSecretValue
                            else:
                                # generate typed SecretValue
                                field_type = Annotated[field_type, SecretSentinel]
                    # remove sentinel from default
                    p = p.replace(default=None)
                elif field_type is AnyType:
                    # try to get type from default
                    field_type = type(p.default)

                if include_defaults or type_from_literal is not None:
                    # set annotations
                    annotations[p.name] = field_type
                    # set field with default value
                    new_fields[p.name] = p.default
                    # print(
                    #     f"Param {p.name} is {field_type}: {p.default} due to {include_defaults} or"
                    #     f" {type_from_literal}"
                    # )

    signature_fields = {**sig_base_fields, **new_fields}

    # new type goes to the module where sig was declared
    new_fields["__module__"] = module.__name__
    # set annotations so they are present in __dict__
    new_fields["__annotations__"] = annotations
    # synthesize type
    T: Type[BaseConfiguration] = type(name, (base,), new_fields)
    SPEC = configspec()(T)
    # add to the module
    setattr(module, name, SPEC)

    return SPEC, signature_fields



import dataclasses
from typing import Final, ClassVar, Any, List, Dict, Optional

from dlt.common.configuration import configspec
from dlt.common.configuration.specs import ConnectionStringCredentials
from dlt.common.utils import digest128
from dlt.common.typing import TSecretStrValue
from dlt.common.exceptions import SystemConfigurationException

from dlt.common.destination.client import DestinationClientDwhWithStagingConfiguration


def escape_mssql_odbc_value(value: Optional[str]) -> str:
    """Escape a value for MSSQL ADO/ODBC connection string format.

    ODBC format supports `{value}` syntax where:
      - `}` inside braces must be doubled to `}}`
      - `;` can safely appear inside braces

    To safely handle values with special characters, we use ODBC-style bracing:
    - Values containing `;` or `}` are wrapped in `{}`
    - `}` inside the value is escaped as `}}`

    Args:
        value: The value to escape

    Returns:
        Escaped value safe for use in ADO/ODBC connection string
    """
    if not value:
        return ""
    # if value contains ; or }, use braced syntax with }} escaping
    if ";" in value or "}" in value:
        return "{" + value.replace("}", "}}") + "}"
    return value


def build_odbc_dsn(params: Dict[str, Any]) -> str:
    """Build an ADO/ODBC connection string for MSSQL, escaping values

    Args:
        params: Dictionary of connection parameters

    Returns:
        ADO/ODBC connection string
    """
    return ";".join(
        f"{k}={escape_mssql_odbc_value(str(v))}" for k, v in params.items() if v is not None
    )


@configspec(init=False)
class MsSqlCredentials(ConnectionStringCredentials):
    drivername: Final[str] = dataclasses.field(default="mssql", init=False, repr=False, compare=False)  # type: ignore
    database: str = None
    username: str = None
    password: TSecretStrValue = None
    host: str = None
    port: int = 1433
    connect_timeout: int = 30
    driver: str = None

    __config_gen_annotations__: ClassVar[List[str]] = ["port", "connect_timeout"]

    SUPPORTED_DRIVERS: ClassVar[List[str]] = [
        "ODBC Driver 18 for SQL Server",
        "ODBC Driver 17 for SQL Server",
    ]

    def parse_native_representation(self, native_value: Any) -> None:
        # TODO: Support ODBC connection string or sqlalchemy URL
        super().parse_native_representation(native_value)
        if self.query is not None:
            self.query = {k.lower(): v for k, v in self.query.items()}  # Make case-insensitive.
        self.driver = self.query.get("driver", self.driver)
        self.connect_timeout = int(self.query.get("connect_timeout", self.connect_timeout))

    def on_resolved(self) -> None:
        if self.driver not in self.SUPPORTED_DRIVERS:
            raise SystemConfigurationException(
                f"The specified driver `{self.driver}` is not supported."
                f" Choose one of the supported drivers: {', '.join(self.SUPPORTED_DRIVERS)}."
            )
        self.database = self.database.lower()

    def get_query(self) -> Dict[str, Any]:
        query = dict(super().get_query())
        query["connect_timeout"] = self.connect_timeout
        return query

    def on_partial(self) -> None:
        self.driver = self._get_driver()
        if not self.is_partial():
            self.resolve()

    def _get_driver(self) -> str:
        if self.driver:
            return self.driver

        # Pick a default driver if available
        import pyodbc

        available_drivers = pyodbc.drivers()
        for d in self.SUPPORTED_DRIVERS:
            if d in available_drivers:
                return d
        docs_url = "https://learn.microsoft.com/en-us/sql/connect/odbc/download-odbc-driver-for-sql-server?view=sql-server-ver16"
        raise SystemConfigurationException(
            f"No supported ODBC driver found for MS SQL Server.  See {docs_url} for information on"
            f" how to install the `{self.SUPPORTED_DRIVERS[0]}` on your platform."
        )

    def get_odbc_dsn_dict(self) -> Dict[str, Any]:
        params = {
            "DRIVER": self.driver,
            "SERVER": f"{self.host},{self.port}",
            "DATABASE": self.database,
            "UID": self.username,
            "PWD": self.password,
        }
        if self.query is not None:
            params.update({k.upper(): v for k, v in self.query.items()})
        return params

    def to_odbc_dsn(self) -> str:
        params = self.get_odbc_dsn_dict()
        return build_odbc_dsn(params)


@configspec
class MsSqlClientConfiguration(DestinationClientDwhWithStagingConfiguration):
    destination_type: Final[str] = dataclasses.field(default="mssql", init=False, repr=False, compare=False)  # type: ignore
    credentials: MsSqlCredentials = None

    create_indexes: bool = False
    has_case_sensitive_identifiers: bool = False

    def fingerprint(self) -> str:
        """Returns a fingerprint of host part of a connection string"""
        if self.credentials and self.credentials.host:
            return digest128(self.credentials.host)
        return ""



import os
import logging
from typing import Any, Sequence, Optional, Union
import warnings

from dlt.common import logger
from dlt.common.json import json
from dlt.common.exceptions import MissingDependencyException
from dlt.common.typing import StrAny, TypeAlias, AnyType

from dlt.helpers.dbt.exceptions import (
    DBTProcessingError,
    DBTNodeResult,
    IncrementalSchemaOutOfSyncError,
)


try:
    from dbt.contracts import results as dbt_results
    from dbt.cli.main import dbtRunner
except ImportError:
    raise MissingDependencyException("DBT Core", ["dbt-core"])


def initialize_dbt_logging(level: str, is_json_logging: bool) -> Sequence[str]:
    int_level = logging._nameToLevel[level]

    globs = []
    if int_level <= logging.DEBUG:
        globs = ["--debug"]
    if int_level >= logging.WARNING:
        globs = ["--quiet", "--no-print"]

    # return global parameters to be passed to setup logging

    if is_json_logging:
        return ["--log-format", "json"] + globs
    else:
        return globs


def is_incremental_schema_out_of_sync_error(error: Any) -> bool:
    def _check_single_item(error_: dbt_results.RunResult) -> bool:
        return (
            error_.status == dbt_results.RunStatus.Error
            and "The source and target schemas on this incremental model are out of sync"
            in error_.message
        )

    if isinstance(error, dbt_results.RunResult):
        return _check_single_item(error)
    elif isinstance(error, dbt_results.RunExecutionResult):
        return any(_check_single_item(r) for r in error.results)

    return False


def parse_dbt_execution_results(results: Any) -> Sequence[DBTNodeResult]:
    # run may return RunResult of something different depending on error
    if isinstance(results, dbt_results.ExecutionResult):
        pass
    elif isinstance(results, dbt_results.BaseResult):
        # a single result
        results = [results]
    else:
        logger.warning(f"{type(results)} is unknown and cannot be logged")
        return None

    return [
        DBTNodeResult(res.node.name, res.message, res.execution_time, str(res.status))
        for res in results
        if isinstance(res, dbt_results.NodeResult)
    ]


def run_dbt_command(
    package_path: str,
    command: str,
    profiles_dir: str,
    profile_name: Optional[str] = None,
    global_args: Sequence[str] = None,
    command_args: Sequence[str] = None,
    package_vars: StrAny = None,
) -> Union[Sequence[DBTNodeResult], dbt_results.ExecutionResult]:
    args = ["--profiles-dir", profiles_dir]
    # add profile name if provided
    if profile_name:
        args += ["--profile", profile_name]
    # serialize dbt variables to pass to package
    if package_vars:
        args += ["--vars", json.dumps(package_vars)]
    if command_args:
        args += command_args

    # cwd to package dir
    working_dir = os.getcwd()
    os.chdir(package_path)
    try:
        results: dbt_results.ExecutionResult = None
        success: bool = None
        # dbt uses logbook which does not run on python 10. below is a hack that allows that
        warnings.filterwarnings("ignore", category=DeprecationWarning, module="logbook")
        runner_args = (global_args or []) + [command] + args  # type: ignore

        runner = dbtRunner()
        run_result = runner.invoke(runner_args)
        success = run_result.success
        results = run_result.result

        assert type(success) is bool
        parsed_results = parse_dbt_execution_results(results)
        if not success:
            dbt_exc = DBTProcessingError(command, parsed_results, results)
            # detect incremental model out of sync
            if is_incremental_schema_out_of_sync_error(results):
                raise IncrementalSchemaOutOfSyncError(dbt_exc)
            raise dbt_exc
        return parsed_results or results
    except SystemExit as sys_ex:
        # dbt >= 1.7 uses dbtRunner which handles BaseExceptions internally,
        # so SystemExit should not reach here, but we handle it defensively

        raise DBTProcessingError(command, None, sys_ex)
    finally:
        # go back to working dir
        os.chdir(working_dir)


def init_logging_and_run_dbt_command(
    log_level: str,
    is_json_logging: bool,
    package_path: str,
    command: str,
    profiles_dir: str,
    profile_name: Optional[str] = None,
    command_args: Sequence[str] = None,
    package_vars: StrAny = None,
) -> Union[Sequence[DBTNodeResult], dbt_results.ExecutionResult]:
    # initialize dbt logging, returns global parameters to dbt command
    dbt_global_args = initialize_dbt_logging(log_level, is_json_logging)
    return run_dbt_command(
        package_path,
        command,
        profiles_dir,
        profile_name,
        dbt_global_args,
        command_args,
        package_vars,
    )



from typing import List, TYPE_CHECKING, Optional

import sqlalchemy as sa  # noqa
from sqlalchemy.schema import CreateColumn

from dlt.common import logger

if TYPE_CHECKING:
    from alembic.operations import Operations


class ListBuffer:
    """A partial implementation of string IO to use with alembic.
    SQL statements are stored in a list instead of file/stdio
    """

    def __init__(self) -> None:
        self._buf = ""
        self.sql_lines: List[str] = []

    def write(self, data: str) -> None:
        self._buf += data

    def flush(self) -> None:
        if self._buf:
            self.sql_lines.append(self._buf)
            self._buf = ""


class MigrationMaker:
    def __init__(self, dialect: sa.engine.Dialect) -> None:
        self._buf = ListBuffer()
        self.dialect = dialect

        # Try to create MigrationContext, fallback to None if alembic not available or dialect not supported
        self.ctx = None
        self.ops: Optional["Operations"] = None

        try:
            from alembic.runtime.migration import MigrationContext
            from alembic.operations import Operations
        except ModuleNotFoundError:
            logger.warning(
                f"Alembic not installed, falling back to direct SQL generation for {dialect.name}. "
                "Install the sqlalchemy extra (dlt[sqlalchemy]) for full ALTER TABLE support."
            )
            return

        try:
            self.ctx = MigrationContext(
                dialect,
                None,
                {
                    "as_sql": True,
                    "output_buffer": self._buf,
                    "mssql_batch_separator": None,
                    "oracle_batch_separator": None,
                },
            )
            self.ops = Operations(self.ctx)
        except KeyError:
            # dialect is not supported by alembic
            logger.warning(
                f"Alembic does not support dialect {dialect.name}, falling back to direct SQL"
                " generation."
            )

    def add_column(self, table_name: str, column: sa.Column, schema: str) -> None:
        if self.ops:
            # Use alembic if available
            self.ops.add_column(table_name, column, schema=schema)
        else:
            # Fallback to direct SQLAlchemy for unsupported dialects
            prep = self.dialect.identifier_preparer  # type: ignore[attr-defined]
            fq_table = prep.quote_identifier(table_name)
            if schema:
                fq_table = f"{prep.quote_identifier(schema)}.{fq_table}"

            # Create column DDL
            col_ddl = CreateColumn(column).compile(
                dialect=self.dialect, compile_kwargs={"literal_binds": True}
            )

            # Generate ALTER TABLE statement
            sql_statement = f"ALTER TABLE {fq_table} ADD COLUMN {col_ddl}"
            self._buf.sql_lines.append(sql_statement)

    def consume_statements(self) -> List[str]:
        lines = self._buf.sql_lines[:]
        self._buf.sql_lines.clear()
        return lines



import gzip
import os
import re
import stat
import errno
import shutil
import pathvalidate
from typing import IO, Any, Optional, List, cast
from dlt.common.typing import AnyFun

from dlt.common.utils import encoding_for_mode, uniq_id


FILE_COMPONENT_INVALID_CHARACTERS = re.compile(r"[.%{}]")


class FileStorage:
    def __init__(self, storage_path: str, file_type: str = "t", makedirs: bool = False) -> None:
        # make it absolute path
        self.storage_path = os.path.realpath(storage_path)
        self.file_type = file_type
        if makedirs:
            os.makedirs(storage_path, exist_ok=True)

    def save(self, relative_path: str, data: Any) -> str:
        return self.save_atomic(self.storage_path, relative_path, data, file_type=self.file_type)

    @staticmethod
    def save_atomic(storage_path: str, relative_path: str, data: Any, file_type: str = "t") -> str:
        mode = "w" + file_type
        tmp_path = os.path.join(storage_path, uniq_id(8))
        with open(tmp_path, mode=mode, encoding=encoding_for_mode(mode)) as f:
            f.write(data)
        try:
            dest_path = os.path.join(storage_path, relative_path)
            # os.rename reverts to os.replace on posix. on windows this operation is not atomic!
            os.replace(tmp_path, dest_path)
            return dest_path
        except Exception:
            if os.path.isfile(tmp_path):
                os.remove(tmp_path)
            raise

    @staticmethod
    def move_atomic_to_folder(source_file_path: str, dest_folder_path: str) -> str:
        file_name = os.path.basename(source_file_path)
        dest_file_path = os.path.join(dest_folder_path, file_name)
        return FileStorage.move_atomic_to_file(source_file_path, dest_file_path)

    @staticmethod
    def move_atomic_to_file(source_file_path: str, dest_file_path: str) -> str:
        try:
            os.rename(source_file_path, dest_file_path)
        except OSError:
            # copy to local temp file
            folder_name = os.path.dirname(dest_file_path)
            dest_temp_file = os.path.join(folder_name, uniq_id())
            try:
                shutil.copyfile(source_file_path, dest_temp_file)
                os.rename(dest_temp_file, dest_file_path)
                os.unlink(source_file_path)
            except Exception:
                if os.path.isfile(dest_temp_file):
                    os.remove(dest_temp_file)
                raise
        return dest_file_path

    @staticmethod
    def copy_atomic_to_file(source_file_path: str, dest_file_path: str) -> str:
        folder_name = os.path.dirname(dest_file_path)
        dest_temp_file = os.path.join(folder_name, uniq_id())
        try:
            shutil.copyfile(source_file_path, dest_temp_file)
            os.rename(dest_temp_file, dest_file_path)
        except Exception:
            if os.path.isfile(dest_temp_file):
                os.remove(dest_temp_file)
            raise
        return dest_file_path

    def load(self, relative_path: str) -> Any:
        # raises on file not existing
        with self.open_file(relative_path) as text_file:
            return text_file.read()

    def delete(self, relative_path: str) -> None:
        file_path = self.make_full_path(relative_path)
        if os.path.isfile(file_path):
            os.remove(file_path)
        else:
            raise FileNotFoundError(file_path)

    def delete_folder(
        self, relative_path: str, recursively: bool = False, delete_ro: bool = False
    ) -> None:
        folder_path = self.make_full_path(relative_path)
        if os.path.isdir(folder_path):
            if recursively:
                if delete_ro:
                    del_ro = self.rmtree_del_ro
                else:
                    del_ro = None
                shutil.rmtree(folder_path, onerror=del_ro)
            else:
                os.rmdir(folder_path)
        else:
            raise NotADirectoryError(folder_path)

    def open_file(self, relative_path: str, mode: str = "r") -> IO[Any]:
        if "b" not in mode and "t" not in mode:
            mode = mode + self.file_type
        if "r" in mode:
            return FileStorage.open_zipsafe_ro(self.make_full_path(relative_path), mode)
        return open(self.make_full_path(relative_path), mode, encoding=encoding_for_mode(mode))

    def touch_file(self, relative_path: str) -> None:
        """Touch file, assumes single writer"""
        file_path = self.make_full_path(relative_path)
        try:
            os.utime(file_path, None)
            return
        except OSError as ex:
            if isinstance(ex, FileNotFoundError):
                # File does not exist, create it
                flags = os.O_CREAT | os.O_EXCL | os.O_WRONLY
                fd = None
                try:
                    fd = os.open(file_path, flags, 0o666)
                    return
                finally:
                    if fd is not None:
                        os.close(fd)
            else:
                # utime does not work
                pass

    # def open_temp(self, delete: bool = False, mode: str = "w", file_type: str = None) -> IO[Any]:
    #     mode = mode + file_type or self.file_type
    #     return tempfile.NamedTemporaryFile(
    #         dir=self.storage_path, mode=mode, delete=delete, encoding=encoding_for_mode(mode)
    #     )

    def has_file(self, relative_path: str) -> bool:
        return os.path.isfile(self.make_full_path(relative_path))

    def has_folder(self, relative_path: str) -> bool:
        return os.path.isdir(self.make_full_path(relative_path))

    def list_folder_files(self, relative_path: str, to_root: bool = True) -> List[str]:
        """List all files in `relative_path` folder

        Args:
            relative_path (str): A path to folder, relative to storage root
            to_root (bool, optional): If True returns paths to files in relation to root, if False, returns just file names. Defaults to True.

        Returns:
            List[str]: A list of file names with optional path as per ``to_root`` parameter
        """
        scan_path = self.make_full_path(relative_path)
        if to_root:
            # list files in relative path, returning paths relative to storage root
            return [
                os.path.join(relative_path, e.name) for e in os.scandir(scan_path) if e.is_file()
            ]
        else:
            # or to the folder
            return [e.name for e in os.scandir(scan_path) if e.is_file()]

    def list_folder_dirs(self, relative_path: str, to_root: bool = True) -> List[str]:
        # list content of relative path, returning paths relative to storage root
        scan_path = self.make_full_path(relative_path)
        if to_root:
            # list folders in relative path, returning paths relative to storage root
            return [
                os.path.join(relative_path, e.name) for e in os.scandir(scan_path) if e.is_dir()
            ]
        else:
            # or to the folder
            return [e.name for e in os.scandir(scan_path) if e.is_dir()]

    def create_folder(self, relative_path: str, exists_ok: bool = False) -> None:
        os.makedirs(self.make_full_path(relative_path), exist_ok=exists_ok)

    def link_hard(self, from_relative_path: str, to_relative_path: str) -> None:
        # note: some interesting stuff on links https://lightrun.com/answers/conan-io-conan-research-investigate-symlinks-and-hard-links
        os.link(self.make_full_path(from_relative_path), self.make_full_path(to_relative_path))

    @staticmethod
    def link_hard_with_fallback(external_file_path: str, to_file_path: str) -> None:
        """Try to create a hardlink and fallback to copying when filesystem doesn't support links"""
        try:
            os.link(external_file_path, to_file_path)
        except OSError:
            # Fallback to copy when fs doesn't support links or attempting to make a cross-device link
            FileStorage.copy_atomic_to_file(external_file_path, to_file_path)

    def atomic_rename(self, from_relative_path: str, to_relative_path: str) -> None:
        """Renames a path using os.rename which is atomic on POSIX, Windows and NFS v4.

        Method falls back to non-atomic method in following cases:
        1. On Windows when destination file exists
        2. If underlying file system does not support atomic rename
        3. All buckets mapped with FUSE are not atomic
        """

        os.rename(self.make_full_path(from_relative_path), self.make_full_path(to_relative_path))

    def rename_tree(self, from_relative_path: str, to_relative_path: str) -> None:
        """Renames a tree using os.rename if possible making it atomic

        If we get 'too many open files': in that case `rename_tree_files is used
        """

        try:
            self.atomic_rename(from_relative_path, to_relative_path)
            return
        except OSError as ex:
            if ex.errno != errno.EMFILE:
                raise
        self.rename_tree_files(from_relative_path, to_relative_path)

    def rename_tree_files(self, from_relative_path: str, to_relative_path: str) -> None:
        """Renames files in a tree recursively using os.rename."""
        from_path = self.make_full_path(from_relative_path)
        to_path = self.make_full_path(to_relative_path)

        if not os.path.isdir(from_path):
            raise ValueError(f"`{from_path}` is not a directory")

        # make sure the destination directory exists
        os.makedirs(to_path, exist_ok=True)

        # move files and directories from source to destination starting from leafs
        for root, _, files in os.walk(from_path, topdown=False):
            to_root = root.replace(from_path, to_path, 1)
            os.makedirs(to_root, exist_ok=True)

            for name in files:
                file_from_path = os.path.join(root, name)
                os.rename(file_from_path, os.path.join(to_root, name))

            if not os.listdir(root):
                os.rmdir(root)

    def atomic_import(
        self, external_file_path: str, to_folder: str, new_file_name: Optional[str] = None
    ) -> str:
        """Moves a file at `external_file_path` into the `to_folder` effectively importing file into storage

        Args:
            external_file_path: Path to file to be imported
            to_folder: Path to folder where file should be imported
            new_file_name: Optional new file name for the imported file, otherwise the original file name is used

        Returns:
            Path to imported file relative to storage root
        """
        new_file_name = new_file_name or os.path.basename(external_file_path)
        dest_file_path = os.path.join(self.make_full_path(to_folder), new_file_name)
        return self.to_relative_path(
            FileStorage.move_atomic_to_file(external_file_path, dest_file_path)
        )

    def is_path_in_storage(self, path: str) -> bool:
        """Checks if a given path is below storage root, without checking for item existence"""
        assert path is not None
        # all paths are relative to root
        if not os.path.isabs(path):
            path = os.path.join(self.storage_path, path)
        file = os.path.realpath(path)
        try:
            return os.path.commonpath([file, self.storage_path]) == self.storage_path
        except ValueError:
            # paths on different drives on windows
            return False

    def to_relative_path(self, path: str) -> str:
        if path == "":
            return ""
        if not self.is_path_in_storage(path):
            raise ValueError(path)
        if not os.path.isabs(path):
            path = os.path.realpath(os.path.join(self.storage_path, path))
        # for abs paths find the relative
        return os.path.relpath(path, start=self.storage_path)

    def make_full_path_safe(self, path: str) -> str:
        """Verifies that path is under storage root and then returns normalized absolute path"""
        # try to make a relative path if paths are absolute or overlapping
        path = self.to_relative_path(path)
        # then assume that it is a path relative to storage root
        return os.path.realpath(os.path.join(self.storage_path, path))

    def make_full_path(self, path: str) -> str:
        """Joins path with storage root. Intended for path known to be relative to storage root"""
        return os.path.join(self.storage_path, path)

    def from_wd_to_relative_path(self, wd_relative_path: str) -> str:
        path = os.path.realpath(wd_relative_path)
        return self.to_relative_path(path)

    def from_relative_path_to_wd(self, relative_path: str) -> str:
        return os.path.relpath(self.make_full_path_safe(relative_path), start=".")

    @staticmethod
    def get_file_name_from_file_path(file_path: str) -> str:
        return os.path.basename(file_path)

    @staticmethod
    def validate_file_name_component(name: str) -> None:
        # Universal platform bans several characters allowed in POSIX ie. | < \ or "COM1" :)
        try:
            pathvalidate.validate_filename(name, platform="Universal")
        except pathvalidate.error.ValidationError as val_ex:
            if val_ex.reason != pathvalidate.ErrorReason.INVALID_LENGTH:
                raise
        # component cannot contain "."
        if FILE_COMPONENT_INVALID_CHARACTERS.search(name):
            raise pathvalidate.error.InvalidCharError(
                description="Component name cannot contain the following characters: . % { }"
            )

    @staticmethod
    def rmtree_del_ro(action: AnyFun, name: str, exc: Any) -> Any:
        if action in (os.unlink, os.remove, os.rmdir):
            os.chmod(name, stat.S_IWRITE)
            action(name)

    @staticmethod
    def open_zipsafe_ro(path: str, mode: str = "r", **kwargs: Any) -> IO[Any]:
        """Opens a file using gzip.open if it is a gzip file, otherwise uses open."""
        assert "r" in mode, "FileStorage.open_zipsafe_ro only supports read modes"
        encoding = kwargs.pop("encoding", encoding_for_mode(mode))
        origmode = str(mode)
        try:
            if encoding is not None and mode == "r":
                mode += "t"  # gzip requires text mode explicitly to use encoding
            f = gzip.open(path, mode, encoding=encoding, **kwargs)
            # Force gzip to read the first few bytes and check the magic number
            f.read(2), f.seek(0)
            return cast(IO[Any], f)
        except (gzip.BadGzipFile, OSError):
            return open(path, origmode, encoding=encoding, **kwargs)

    @staticmethod
    def is_gzipped(path: str) -> bool:
        """Checks if file under path is gzipped by reading a header"""
        try:
            with gzip.open(path, "rt", encoding="utf-8") as f:
                # Force gzip to read the first few bytes and check the magic number
                f.read(2)
                return True
        except (gzip.BadGzipFile, OSError):
            return False



import os
import threading
import warnings
from typing import TYPE_CHECKING, ClassVar, List, Optional, Any, Final, Type, Dict, Union
import dataclasses

from dlt.common import logger
from dlt.common.configuration import configspec
from dlt.common.configuration.specs import ConnectionStringCredentials
from dlt.common.configuration.specs.base_configuration import NotResolved
from dlt.common.destination.client import DestinationClientDwhConfiguration
from dlt.common.storages.configuration import WithLocalFiles
from dlt.common.typing import Annotated
from dlt.common.warnings import DltDeprecationWarning

if TYPE_CHECKING:
    from sqlalchemy.engine import Engine, Dialect, Connection


class ManagedEngine:
    """Thread-safe engine lifecycle with ref-counted connection borrowing.

    Owned engines (created from URL) are disposed when the last borrowed
    connection is returned. External engines are never disposed.
    """

    def __init__(
        self,
        credentials: "SqlalchemyCredentials",
        engine: Optional["Engine"] = None,
    ) -> None:
        self._credentials = credentials
        self._conn_lock = threading.RLock()
        self._conn_borrows: int = 0
        if engine is not None:
            # external engine — not owned, not disposed by us
            self._engine: Optional["Engine"] = engine
            self._conn_owner: bool = False
        else:
            # owned engine — lazily created, disposed when refcount hits 0
            self._engine = None
            self._conn_owner = True

    @property
    def engine(self) -> "Engine":
        """Lazily creates the engine for owned instances, returns external engine otherwise."""
        from dlt.common.libs.sql_alchemy import sa

        if self._engine is None:
            engine_kwargs = self._credentials._resolve_engine_kwargs()
            self._engine = sa.create_engine(
                self._credentials.to_url().render_as_string(hide_password=False),
                **engine_kwargs,
            )
        return self._engine

    def borrow_conn(self) -> "Connection":
        """Borrow a connection from the engine. Must be returned via return_conn."""
        with self._conn_lock:
            engine_ = self.engine
            self._conn_borrows += 1
        try:
            return engine_.connect()
        except Exception:
            with self._conn_lock:
                self._conn_borrows -= 1
                if self._conn_borrows == 0 and self._conn_owner:
                    self._dispose_engine()
            raise

    def return_conn(self, borrowed_conn: "Connection") -> None:
        """Return a borrowed connection. Disposes owned engine when refcount hits 0."""
        with self._conn_lock:
            borrowed_conn.close()
            assert self._conn_borrows > 0, "Returning connection when borrows is 0"
            self._conn_borrows -= 1
            if self._conn_borrows == 0 and self._conn_owner:
                self._dispose_engine()

    def _dispose_engine(self) -> None:
        if self._conn_borrows > 0:
            warnings.warn(
                f"Disposing engine {self._engine.url} with {self._conn_borrows} open conns."
            )
        if self._engine:
            self._engine.dispose()
            self._engine = None

    def __del__(self) -> None:
        if getattr(self, "_engine", None) and getattr(self, "_conn_owner", False):
            self._dispose_engine()


@configspec(init=False)
class SqlalchemyCredentials(ConnectionStringCredentials):
    engine_kwargs: Optional[Dict[str, Any]] = None
    """Additional keyword arguments passed to `sqlalchemy.create_engine`"""

    engine_args: Optional[Dict[str, Any]] = None
    """DEPRECATED: use engine_kwargs instead"""

    managed_engine: Annotated[Optional[ManagedEngine], NotResolved()] = None

    def __init__(
        self, connection_string: Optional[Union[str, Dict[str, Any], "Engine"]] = None
    ) -> None:
        super().__init__(connection_string)  # type: ignore[arg-type]

    def copy(self: "SqlalchemyCredentials") -> "SqlalchemyCredentials":
        new_obj = super().copy()
        # copy always holds the engine as unmanaged (never disposes)
        if self.managed_engine is not None and self.managed_engine._engine is not None:
            new_obj.managed_engine = ManagedEngine(new_obj, engine=self.managed_engine._engine)
        else:
            new_obj.managed_engine = None
        return new_obj

    def _resolve_engine_kwargs(self) -> Dict[str, Any]:
        if self.engine_kwargs and self.engine_args:
            raise ValueError(
                "Both engine_kwargs and engine_args were provided. Use engine_kwargs only."
            )

        if self.engine_args and not self.engine_kwargs:
            warnings.warn(
                DltDeprecationWarning(
                    "`engine_args` is deprecated; use `engine_kwargs` instead",
                    since="1.21.0",
                ),
                stacklevel=2,
            )
            return self.engine_args

        return self.engine_kwargs or {}

    def parse_native_representation(self, native_value: Any) -> None:
        from sqlalchemy.engine import Engine

        if isinstance(native_value, Engine):
            # triggers setter which creates ManagedEngine wrapping external engine
            self.engine = native_value
            super().parse_native_representation(
                native_value.url.render_as_string(hide_password=False)
            )
        else:
            super().parse_native_representation(native_value)

    def _ensure_managed_engine(self) -> ManagedEngine:
        """Lazily create ManagedEngine on first access."""
        if self.managed_engine is None:
            external = getattr(self, "_external_engine", None)
            self.managed_engine = ManagedEngine(self, engine=external)
        return self.managed_engine

    @property
    def engine(self) -> Optional["Engine"]:
        return self._ensure_managed_engine().engine

    @engine.setter
    def engine(self, value: "Engine") -> None:
        # backward compat: create ManagedEngine wrapping external engine
        self._external_engine = value
        self.managed_engine = ManagedEngine(self, engine=value)

    def get_dialect_class(self) -> Optional[Type["Dialect"]]:
        if not self.drivername:
            return None
        if engine := self.engine:
            return type(engine.dialect)
        return self.to_url().get_dialect()  # type: ignore[attr-defined,no-any-return]

    def get_backend_name(self) -> str:
        if not self.drivername:
            return None
        return self.to_url().get_backend_name()

    @staticmethod
    def is_memory_database(database: Optional[str], query: Optional[Dict[str, Any]] = None) -> bool:
        """Detect if the given database/query represent an in-memory SQLite database.

        Handles the following forms:
        - Classic: ``:memory:``
        - URI: ``file:<name>?mode=memory&cache=shared&uri=true``

        NOTE: empty/None database is NOT considered in-memory here because dlt's
        configuration system resolves it to a default file path.
        """
        if database == ":memory:":
            return True
        # URI format: file:<name> with mode=memory and uri=true in query params
        # uri=true is required for pysqlite to interpret the database as a URI
        if (
            database
            and database.startswith("file:")
            and query
            and query.get("mode") == "memory"
            and str(query.get("uri", "")).lower() == "true"
        ):
            return True
        return False

    __config_gen_annotations__: ClassVar[List[str]] = [
        "database",
        "port",
        "username",
        "password",
        "host",
    ]


SQLITE_DB_NAME_PAT = "%s.db"


@configspec
class SqlalchemyClientConfiguration(WithLocalFiles, DestinationClientDwhConfiguration):
    destination_type: Final[str] = dataclasses.field(default="sqlalchemy", init=False, repr=False, compare=False)  # type: ignore
    credentials: SqlalchemyCredentials = None
    """SQLAlchemy connection string"""
    create_unique_indexes: bool = False
    """Whether UNIQUE constrains should be created"""
    create_primary_keys: bool = False
    """Whether PRIMARY KEY constrains should be created"""

    engine_kwargs: Dict[str, Any] = dataclasses.field(default_factory=dict)
    """Additional keyword arguments passed to `sqlalchemy.create_engine`"""

    engine_args: Dict[str, Any] = dataclasses.field(default_factory=dict)
    """DEPRECATED: use engine_kwargs instead"""

    def get_dialect_class(self) -> Type["Dialect"]:
        return self.credentials.get_dialect_class()

    def get_backend_name(self) -> str:
        return self.credentials.get_backend_name()

    def on_resolved(self) -> None:
        if self.engine_kwargs and self.engine_args:
            raise ValueError(
                "Both engine_kwargs and engine_args were provided. Use engine_kwargs only."
            )

        if self.engine_args and not self.engine_kwargs:
            warnings.warn(
                DltDeprecationWarning(
                    "`engine_args` is deprecated; use `engine_kwargs` instead",
                    since="1.21.0",
                ),
                stacklevel=2,
            )
            self.engine_kwargs = self.engine_args

        if self.engine_kwargs and not self.credentials.engine_kwargs:
            self.credentials.engine_kwargs = self.engine_kwargs

        # resolve local file path for sqlite backends
        if self.credentials.drivername == "sqlite":
            if not SqlalchemyCredentials.is_memory_database(
                self.credentials.database, self.credentials.query
            ):
                db = self.credentials.database
                if not db or not os.path.isabs(db):
                    self.credentials.database = os.path.normpath(
                        self.make_location(db or None, SQLITE_DB_NAME_PAT)
                    )
